# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for dot_general."""
import copy
import functools
from typing import Callable

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax.v2 import aqt_quantizer
from aqt.jax.v2 import config
from aqt.jax.v2 import stochastic_rounding
import aqt.jax.v2.aqt_conv_general as aqt_conv
import aqt.jax.v2.aqt_dot_general as aqt
from aqt.jax.v2.numerics import int_numerics
from aqt.jax.v2.numerics import no_numerics
from aqt.jax.v2.numerics import numerics
import flax.linen.linear as fl
import flax.struct
import jax
import jax.numpy as jnp
import numpy as np
import scipy.stats


def test_jaxpr_dtype(f, cfgs: list[aqt.DotGeneralRaw], float_dtype):
  """Tests whether dot_generals in f conform to dtypes inside of cfgs."""

  def jaxpr_to_trityp(jaxpr):
    for eq in jaxpr.eqns:
      # This is where dot_generals hide when one uses custom vjp
      if "fun_jaxpr" in eq.params.keys():
        for rec in jaxpr_to_trityp(eq.params["fun_jaxpr"]):
          yield rec

      if eq.primitive.name == "dot_general":
        [lhs, rhs] = eq.invars
        [out] = eq.outvars
        trityp = (lhs.aval, rhs.aval, out.aval)
        yield trityp

  f_jaxpr = jax.make_jaxpr(f)()
  trityps = [trityp for trityp in jaxpr_to_trityp(f_jaxpr)]
  assert len(trityps) == len(cfgs)
  for (lhs_sa, rhs_sa, out_sa), cfg in zip(trityps, cfgs):
    # If cfg has None, the type is inherited from the arguments' type.
    def assert_dtype_eq(dtype1, dtype2):
      assert dtype1 == dtype2, f"dtype1 != dtype2: {dtype1=} != {dtype2=}"

    assert_dtype_eq(
        lhs_sa.dtype, cfg.lhs.quantizer.numerics.get_dtype() or float_dtype
    )
    assert_dtype_eq(
        rhs_sa.dtype, cfg.rhs.quantizer.numerics.get_dtype() or float_dtype
    )
    assert_dtype_eq(out_sa.dtype, cfg.dg_accumulator_dtype or float_dtype)


def rand_unif(shape, maxval, seed, dtype=jnp.float32):
  key = jax.random.PRNGKey(seed)
  return jax.random.uniform(
      key=key, shape=shape, minval=-maxval, maxval=maxval, dtype=dtype
  )


# The main test strategy is :
# - Test that FQ is sensible (improve fq_noise_test)
#   - Quantization noise (rounding error) is of proper value
#   - Cliping is correct
#   - Stochastic noise is added correctly.
# - Test that we really use int8/int4/bool in XLA/HLO
# - Test of fq vs dot_general equivalence in po2
# - Test that it all applies to gradients
#
#
# TODO(lew): Tests are a bit incomplete. What we need:
# On top of that we shell test:
#  - Gradinent is 1 in the range, 0 outside the abs-max range.


def test_eq(name, a, b):
  # TODO(lew): use library function.
  mean_err = jnp.mean(jnp.abs(a - b))
  if mean_err != 0.0:
    print("mean_err =", mean_err)
    print(a.shape)
    print(a[:3, :3])
    print(b.shape)
    print(b[:3, :3])
    print(f"FAIL: {name}")
    assert False


# Test exact equalities.
def _check_result_eq(dgs, *, lhs, rhs, gra):
  good_lr, good_gl, good_gr = None, None, None
  for name, dg, options in dgs:
    # Default settings:
    if "test_gradient" not in options:
      options["test_gradient"] = True
    if "mult" not in options:
      options["mult"] = (1.0, 1.0, 1.0)
    if "check_fwd_lhs_tricky_clip_and_round" not in options:
      options["check_fwd_lhs_tricky_clip_and_round"] = False

    lrf = dg(lhs, rhs)
    lr, backprop = jax.vjp(dg, lhs, rhs)

    gl, gr = backprop(gra) if options["test_gradient"] else (None, None)

    if options["check_fwd_lhs_tricky_clip_and_round"]:
      # Test that all the expected were zero anyway
      where_zero_gradient_expected = lhs < 0
      assert (gl[where_zero_gradient_expected] == 0.0).all()

    good_lr = lr if good_lr is None else good_lr
    good_gl = gl if good_gl is None else good_gl
    good_gr = gr if good_gr is None else good_gr

    test_eq(f"{name}: lr vs lrf", lr, lrf)

    lr_mult, gl_mult, gr_mult = options["mult"]
    test_eq(f"{name}: lr", good_lr, lr / lr_mult)  # forward pass
    if options["test_gradient"]:
      test_eq(f"{name}: gl", good_gl, gl / gl_mult)  # backward pass  # pytype: disable=unsupported-operands
      test_eq(f"{name}: gr", good_gr, gr / gr_mult)  # backward pass  # pytype: disable=unsupported-operands


def fqt_param_dict(s, use_fwd_quant, **kwargs):
  return dict(
      cfg=config.fully_quantized(
          use_fwd_quant=use_fwd_quant,
          use_stochastic_rounding=False,
          **kwargs,
      ),
      seed=s,
  )


class _TrickyNumerics(numerics.AqtNumerics, flax.struct.PyTreeNode):
  # Needed because int8 casting would do additional clip and round.
  dtype: jnp.dtype | None = None

  def get_dtype(self):
    return self.dtype

  def abs_val_mapped_to(self) -> jnp.ndarray:
    return jnp.array(1.0)

  def fwd(self, x, context):
    del context
    return jax.lax.round(x, jax.lax.RoundingMethod.TO_NEAREST_EVEN)

  def vjp_fwd(self, x, context):
    res = (x,)
    return self.fwd(x, context), res

  def vjp_bwd(self, res, grad):
    (x,) = res
    ret = grad * (x >= 0)
    return (ret, None)


def _modify_cfg(
    readonly_cfg: aqt.DotGeneral,
    *,
    lhs_dequant_mode: aqt.DequantMode = aqt.DequantMode.OUTPUT,
    rhs_dequant_mode: aqt.DequantMode = aqt.DequantMode.OUTPUT,
    lhs_calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    rhs_calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    use_fwd_quant: bool | None = None,
    fwd_lhs_tricky_clip_and_round: bool = False,
    local_aqt: aqt.LocalAqt | None = None,
    clip_gradient: bool = False,
) -> aqt.DotGeneral:
  cfg = copy.deepcopy(readonly_cfg)
  if fwd_lhs_tricky_clip_and_round:
    # Tricky means that we have zero gradient on x < 0
    cfg.fwd.lhs.quantizer.numerics = _TrickyNumerics()
    cfg.fwd.dg_accumulator_dtype = None

  # Setting po2_scale is ensuring that fake_quant and full dot_general
  # have the same numerics when scales are power of two (po2).
  # We are passing dims to config so that we can reuse it in fake_quant.
  # Power-of-2 scales allow FQ and AQT to be exactly the same.
  def _apply_po2_scale(c):
    c.lhs.quantizer.po2_scale = True
    c.rhs.quantizer.po2_scale = True

  def _apply_dequant_mode(c, lhs_dequant_mode, rhs_dequant_mode):
    c.lhs.dequant_mode = lhs_dequant_mode
    c.rhs.dequant_mode = rhs_dequant_mode

  def _apply_calibration_mode(
      c, lhs_calibration_mode, rhs_calibration_mode
  ):
    c.lhs.calibration_mode = lhs_calibration_mode
    c.rhs.calibration_mode = rhs_calibration_mode

  def _disable_quant_types(c, on_lhs=True, on_rhs=True):
    if on_lhs:
      c.lhs.quantizer.numerics = c.lhs.quantizer.numerics.replace(dtype=None)
    if on_rhs:
      c.rhs.quantizer.numerics = c.rhs.quantizer.numerics.replace(dtype=None)
    if on_lhs or on_rhs:
      c.dg_accumulator_dtype = None

  disable_lhs_quant = lhs_dequant_mode == aqt.DequantMode.THIS_INPUT
  disable_rhs_quant = rhs_dequant_mode == aqt.DequantMode.THIS_INPUT
  for c in [cfg.fwd, cfg.dlhs, cfg.drhs]:
    _apply_po2_scale(c)
    _apply_dequant_mode(c, lhs_dequant_mode, rhs_dequant_mode)
    _apply_calibration_mode(
        c, lhs_calibration_mode, rhs_calibration_mode
    )
    _disable_quant_types(c, disable_lhs_quant, disable_rhs_quant)

  if use_fwd_quant is not None:
    # If we disable all rounding, we are just testing whether the scales are
    # correct. We don't even need to disable clipping and we are testing
    # that the scales are not too large.
    def disable_quant(c):
      _disable_quant_types(c)
      if isinstance(c.lhs.quantizer.numerics, int_numerics.IntNumerics):
        c.lhs.quantizer.numerics = c.lhs.quantizer.numerics.replace(round=False)
      if isinstance(c.rhs.quantizer.numerics, int_numerics.IntNumerics):
        c.rhs.quantizer.numerics = c.rhs.quantizer.numerics.replace(round=False)
      # c.lhs.quantizer.numerics.clip = False
      # c.rhs.quantizer.numerics.clip = False

    disable_quant(cfg.fwd)
    disable_quant(cfg.dlhs)
    disable_quant(cfg.drhs)
    lhs_quant = not isinstance(
        cfg.fwd.lhs.quantizer.numerics, no_numerics.NoNumerics
    )
    rhs_quant = not isinstance(
        cfg.fwd.rhs.quantizer.numerics, no_numerics.NoNumerics
    )
    if lhs_quant:
      cfg.drhs.rhs.use_fwd_quant = use_fwd_quant
    if rhs_quant:
      cfg.dlhs.rhs.use_fwd_quant = use_fwd_quant
  if local_aqt is not None:
    # Currently we are not supporting local_aqt in fwd pass
    # cfg.fwd.local_aqt = local_aqt
    cfg.dlhs.local_aqt = local_aqt
    cfg.drhs.local_aqt = local_aqt

    # When using abs-max scaling, this should be a no-op.
  if isinstance(cfg.fwd.lhs.quantizer.numerics, int_numerics.IntNumerics):
    cfg.fwd.lhs.quantizer.numerics = cfg.fwd.lhs.quantizer.numerics.replace(
        clip_gradient=clip_gradient
    )
  if isinstance(cfg.fwd.rhs.quantizer.numerics, int_numerics.IntNumerics):
    cfg.fwd.rhs.quantizer.numerics = cfg.fwd.rhs.quantizer.numerics.replace(
        clip_gradient=clip_gradient
    )

  return cfg


def _aqt_dg_full_lr_diff(
    lhs_dequant_mode: aqt.DequantMode,
    rhs_dequant_mode: aqt.DequantMode,
    lhs_calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    rhs_calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    use_fwd_quant: bool | None = None,
    fwd_lhs_tricky_clip_and_round: bool = False,
    local_aqt: aqt.LocalAqt | None = None,
    *,
    readonly_cfg: aqt.DotGeneral,
    dims: jax.lax.DotDimensionNumbers,
    clip_gradient: bool = False,
) -> Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray]:
  cfg = _modify_cfg(
      readonly_cfg,
      lhs_dequant_mode=lhs_dequant_mode,
      rhs_dequant_mode=rhs_dequant_mode,
      lhs_calibration_mode=lhs_calibration_mode,
      rhs_calibration_mode=rhs_calibration_mode,
      use_fwd_quant=use_fwd_quant,
      fwd_lhs_tricky_clip_and_round=fwd_lhs_tricky_clip_and_round,
      local_aqt=local_aqt,
      clip_gradient=clip_gradient,
  )
  cfg = config.set_context(cfg, key=jax.random.PRNGKey(4), train_step=None)
  dg = aqt.make_dot_general(cfg)
  return lambda lhs, rhs: dg(lhs, rhs, dims)


def _aqt_dg_full(
    dequant_mode: aqt.DequantMode,
    calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    use_fwd_quant: bool | None = None,
    fwd_lhs_tricky_clip_and_round: bool = False,
    local_aqt: aqt.LocalAqt | None = None,
    *,
    readonly_cfg: aqt.DotGeneral,
    dims: jax.lax.DotDimensionNumbers,
    clip_gradient: bool = False,
) -> Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray]:
  return _aqt_dg_full_lr_diff(
      dequant_mode,
      dequant_mode,
      calibration_mode,
      calibration_mode,
      use_fwd_quant,
      fwd_lhs_tricky_clip_and_round,
      local_aqt,
      readonly_cfg=readonly_cfg,
      dims=dims,
      clip_gradient=clip_gradient
  )


def _aqt_dg_raw_lr_diff(
    lhs_dequant_mode: aqt.DequantMode,
    rhs_dequant_mode: aqt.DequantMode,
    lhs_calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    rhs_calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    *,
    readonly_cfg: aqt.DotGeneral,
    dims: jax.lax.DotDimensionNumbers,
) -> Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray]:
  cfg = _modify_cfg(
      readonly_cfg,
      lhs_dequant_mode=lhs_dequant_mode,
      rhs_dequant_mode=rhs_dequant_mode,
      lhs_calibration_mode=lhs_calibration_mode,
      rhs_calibration_mode=rhs_calibration_mode,
  )
  cfg = config.set_context(cfg, key=jax.random.PRNGKey(4), train_step=None)
  return lambda lhs, rhs: cfg.fwd(lhs, rhs, None, None, dims)[0]


def _aqt_dg_raw(
    dequant_mode: aqt.DequantMode,
    calibration_mode: aqt.CalibrationMode = aqt.CalibrationMode.CONTRACTING_AXIS,
    *,
    readonly_cfg: aqt.DotGeneral,
    dims: jax.lax.DotDimensionNumbers,
) -> Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray]:
  return _aqt_dg_raw_lr_diff(
      dequant_mode,
      dequant_mode,
      calibration_mode,
      calibration_mode,
      readonly_cfg=readonly_cfg,
      dims=dims,
  )


class AqtDotGeneralResearchTest(parameterized.TestCase):

  def test_empty(self):
    t = np.random.normal(size=6).reshape((2, 3))
    np.testing.assert_array_equal(t, t)

  @parameterized.product(
      preserve_zero=[True, False],
      prec=[1, 2, 4, 8],
      v=[0.1, 1000.0],
      seed=list(range(10)),
  )
  def test_fq_noise(self, preserve_zero, prec, v, seed):
    key = jax.random.PRNGKey(seed)
    cfg = config.tensor_make(prec)
    if isinstance(cfg.quantizer.numerics, int_numerics.IntNumerics):
      cfg.quantizer.numerics.preserve_zero = preserve_zero
      if not preserve_zero:
        cfg.quantizer.numerics.dtype = None
    cfg.quantizer.calib_shared_axes = (0,)
    sample_size = 10000
    shape = (sample_size,)
    a = jax.random.uniform(key, shape, minval=-v, maxval=v)
    a_fq = aqt_quantizer.make_fake_quant(cfg)(a)
    bucket_noise = a_fq - a  #  ~ U(-bucket_size/2, bucket_size/2)
    bucket_count = (2**prec - 1) if preserve_zero else (2**prec)
    bucket_size = (v * 2) / bucket_count
    noise = bucket_noise / bucket_size + 0.5  # ~U(0, 1)
    pvalue = scipy.stats.kstest(noise, "uniform").pvalue
    assert pvalue > 0.01

  def test_stochastic_rounding_noise(self):
    repeats = 1000
    num_values = 100000
    shape = (repeats, num_values)

    def assert_clt(noise: jnp.ndarray):
      # Test if the sample mean of the noise conforms to Central Limit Theorem

      # std of a uniform distribution [-0.5, 0.5]
      noise_std = 1 / jnp.sqrt(12.0)
      noise_mean = jnp.mean(noise, axis=1)
      assert jnp.all(noise_mean * jnp.sqrt(num_values) < 4 * noise_std)

    # jax.uniform implementation
    noise_fn = stochastic_rounding.JaxUniform()
    jax_uniform_noise = noise_fn(shape, jax.random.PRNGKey(10))
    assert_clt(jax_uniform_noise)
    # customized more efficient implementation
    noise_fn = stochastic_rounding.RandomCenteredUniform()
    custom_1_noise = noise_fn(shape, jax.random.PRNGKey(11))
    assert_clt(custom_1_noise)

  @parameterized.parameters([
      dict(bits=1),
  ])
  def test_fake_quant(
      self,
      bits=4,
      maxval=10.0,
      shape=(20, 1),
  ):
    cfg = config.tensor_make(bits)
    cfg.quantizer.po2_scale = True
    cfg.quantizer.calib_shared_axes = (0,)
    x = jnp.linspace(-maxval, maxval, num=shape[0]).reshape(shape)
    grad = jnp.ones(shape) * 12345.0
    x_fq, backprop = jax.vjp(aqt_quantizer.make_fake_quant(cfg), x)
    gx_fq = backprop(grad)
    # print(f"x     =\n{x}")
    # print(f"x_fq  =\n{x_fq}")
    # print(f"grad  =\n{grad}")
    # print(f"gx_fq =\n{gx_fq}")
    del x
    del x_fq
    del grad
    del gx_fq

    # TODO(lew): test

  @parameterized.parameters([
      dict(
          # TODO(aqt): Change dlhs_bits to 4bit once
          # https://github.com/google/jax/issues/19682 is fixed.
          cfg=config.config_v3(
              fwd_bits=3,
              dlhs_bits=6,
              drhs_bits=5,
              drhs_accumulator_dtype=jnp.int32,  # overwriting the default None
          )
      ),
      dict(cfg=config.dot_general_make(None, None)),
      dict(cfg=config.dot_general_make(1, 1)),
      dict(cfg=config.dot_general_make(1, 2)),
      dict(cfg=config.dot_general_make(2, 1)),
      dict(cfg=config.dot_general_make(2, 2)),
      dict(cfg=config.dot_general_make(8, 8)),
      dict(cfg=config.dot_general_make(8, 8), clip_gradient=True),
      dict(cfg=config.dot_general_make(8, 8, dlhs_local_aqt=aqt.LocalAqt(2))),
      dict(cfg=config.dot_general_make(8, 8, drhs_local_aqt=aqt.LocalAqt(2))),
      # That test could fail numerically because bf16
      # can't keep in the product of int8*int8 accurately.
      # It just so happens that this test does not fail but others do.
      # We do this test anyway, to catch jax-compilation-time errors.
      dict(cfg=config.dot_general_make(2, 2), dtype=jnp.bfloat16),
      dict(cfg=config.dot_general_make(8, 8), dtype=jnp.bfloat16),
      dict(cfg=config.dot_general_make(None, 8)),
      dict(cfg=config.dot_general_make(8, None)),
      dict(
          cfg=fqt_param_dict(s=10, use_fwd_quant=True)["cfg"],
          dims=(((0, 2), (1, 0)), ((3, 1), (2, 4))),
          # contraction: 2, 5; batch: 4, 3
          lhs_shape=(2, 3, 5, 4),  # non-contr: 3, 4
          rhs_shape=(5, 2, 4, 6, 3),  # non-contr: 4, 6, 3
          gra_shape=(4, 3, 6),
      ),
      dict(
          cfg=fqt_param_dict(
              s=10,
              use_fwd_quant=True,
              dlhs_local_aqt=aqt.LocalAqt(2),
          )["cfg"],
          dims=(((0, 2), (1, 0)), ((3, 1), (2, 4))),
          # contraction: 2, 5; batch: 4, 3
          lhs_shape=(2, 3, 5, 4),  # non-contr: 3, 4
          rhs_shape=(5, 2, 4, 6, 3),  # non-contr: 4, 6, 3
          gra_shape=(4, 3, 6),
      ),
      dict(
          cfg=config.dot_general_make(2, 2),
          dims=(((0, 2), (1, 0)), ((3, 1), (2, 4))),
          # contraction: 2, 5; batch: 4, 3
          lhs_shape=(2, 3, 5, 4),  # non-contr: 3, 4
          rhs_shape=(5, 2, 4, 6, 3),  # non-contr: 4, 6, 3
          gra_shape=(4, 3, 6),
      ),
      *[fqt_param_dict(s, use_fwd_quant=False) for s in range(10)],
      *[fqt_param_dict(s, use_fwd_quant=True) for s in range(10)],
  ])
  def test_dot_general_calibration_with_contracting_axis(
      self,
      cfg: aqt.DotGeneral,
      lhs_maxval=10.0,
      rhs_maxval=20.0,
      gra_maxval=30.0,
      dims=(((1,), (0,)), ((), ())),  # classical matmul
      # lhs_shape=(1, 1),
      # rhs_shape=(1, 2),
      # gra_shape=(1, 2),  # has to be the shape of the output
      # lhs_shape=(1, 2),
      # rhs_shape=(2, 3),
      # gra_shape=(1, 3),  # has to be the shape of the output
      lhs_shape=(10, 20),
      rhs_shape=(20, 30),
      gra_shape=(10, 30),  # has to be the shape of the output
      seed=0,
      dtype=jnp.float32,
      clip_gradient=False,
  ):
    readonly_cfg = cfg
    del cfg

    lhs = rand_unif(lhs_shape, lhs_maxval, seed, dtype)
    rhs = rand_unif(rhs_shape, rhs_maxval, seed + 1, dtype)
    gra = rand_unif(gra_shape, gra_maxval, seed + 2, dtype)

    # Prepare utility functions for test.
    aqt_dg_full = functools.partial(
        _aqt_dg_full,
        readonly_cfg=readonly_cfg,
        dims=dims,
        clip_gradient=clip_gradient,
    )
    aqt_dg_raw = functools.partial(
        _aqt_dg_raw, readonly_cfg=readonly_cfg, dims=dims
    )
    modify_cfg = functools.partial(_modify_cfg, readonly_cfg=readonly_cfg)
    check = functools.partial(_check_result_eq, lhs=lhs, rhs=rhs, gra=gra)

    # Tests for dot_general.
    test_jaxpr_dtype(
        lambda: aqt_dg_full(aqt.DequantMode.OUTPUT)(lhs, rhs),
        [modify_cfg().fwd],
        lhs.dtype,
    )
    test_jaxpr_dtype(
        lambda: jax.vjp(aqt_dg_full(aqt.DequantMode.OUTPUT), lhs, rhs),
        [modify_cfg().fwd],
        lhs.dtype,
    )
    _, backprop = jax.vjp(aqt_dg_full(aqt.DequantMode.OUTPUT), lhs, rhs)
    test_jaxpr_dtype(
        lambda: backprop(gra),
        [modify_cfg().dlhs, modify_cfg().drhs],
        gra.dtype,
    )

    check([
        ("default    ", aqt_dg_full(aqt.DequantMode.OUTPUT), dict()),
        ("FQ         ", aqt_dg_full(aqt.DequantMode.THIS_INPUT), dict()),
        (
            "raw fwd    ",
            aqt_dg_raw(aqt.DequantMode.OUTPUT),
            dict(test_gradient=False),
        ),
        (
            "raw fwd FQ ",
            aqt_dg_raw(aqt.DequantMode.THIS_INPUT),
            dict(test_gradient=False),
        ),
    ])

    check([
        (
            "fwd_quant=T",
            aqt_dg_full(aqt.DequantMode.OUTPUT, use_fwd_quant=False),
            dict(),
        ),
        (
            "fwd_quant=F",
            aqt_dg_full(aqt.DequantMode.OUTPUT, use_fwd_quant=True),
            dict(),
        ),
    ])

    check([
        (
            "default    ",
            aqt_dg_full(aqt.DequantMode.OUTPUT, local_aqt=aqt.LocalAqt(2)),
            dict(),
        ),
        (
            "default    ",
            aqt_dg_full(aqt.DequantMode.THIS_INPUT, local_aqt=aqt.LocalAqt(2)),
            dict(),
        ),
    ])

    if isinstance(
        readonly_cfg.fwd.lhs.quantizer.numerics, int_numerics.IntNumerics
    ):
      check([
          (
              "check_fwd_lhs_tricky_clip_and_round",
              aqt_dg_full(
                  aqt.DequantMode.OUTPUT, fwd_lhs_tricky_clip_and_round=True
              ),
              dict(check_fwd_lhs_tricky_clip_and_round=True),
          ),
      ])

    def unquant_aqt_dg(lhs, rhs):
      dg = aqt.make_dot_general(config.default_unquantized_config())
      return dg(lhs, rhs, dims)

    def lax_dg(lhs, rhs):
      return jax.lax.dot_general(lhs, rhs, dims)

    check([
        ("unquantized default:", unquant_aqt_dg, dict()),
        ("lax.dot_general:", lax_dg, dict()),
    ])

  @parameterized.parameters([
      dict(
          cfg=config.config_v3(
              fwd_bits=3,
              dlhs_bits=4,
              drhs_bits=5,
              drhs_accumulator_dtype=jnp.int32,  # overwriting the default None
          )
      ),
      dict(cfg=config.dot_general_make(None, None)),
      dict(cfg=config.dot_general_make(1, 1)),
      dict(cfg=config.dot_general_make(1, 2)),
      dict(cfg=config.dot_general_make(2, 1)),
      dict(cfg=config.dot_general_make(2, 2)),
      dict(cfg=config.dot_general_make(8, 8)),
      dict(cfg=config.dot_general_make(8, 8), clip_gradient=True),
      dict(cfg=config.dot_general_make(8, 8, dlhs_local_aqt=aqt.LocalAqt(2))),
      dict(cfg=config.dot_general_make(8, 8, drhs_local_aqt=aqt.LocalAqt(2))),
      # That test could fail numerically because bf16
      # can't keep in the product of int8*int8 accurately.
      # It just so happens that this test does not fail but others do.
      # We do this test anyway, to catch jax-compilation-time errors.
      dict(cfg=config.dot_general_make(2, 2), dtype=jnp.bfloat16),
      dict(cfg=config.dot_general_make(8, 8), dtype=jnp.bfloat16),
      dict(cfg=config.dot_general_make(None, 8)),
      dict(cfg=config.dot_general_make(8, None)),
      dict(
          cfg=config.dot_general_make(2, 2),
          dims=(((0, 2), (1, 0)), ((3, 1), (2, 4))),
          # contraction: 2, 5; batch: 4, 3
          lhs_shape=(2, 3, 5, 4),  # non-contr: 3, 4
          rhs_shape=(5, 2, 4, 6, 3),  # non-contr: 4, 6, 3
          gra_shape=(4, 3, 6),
      ),
      *[fqt_param_dict(s, use_fwd_quant=None) for s in range(10)],
  ])
  def test_dot_general_calibration_with_remaining_axis(
      self,
      cfg: config.DotGeneral,
      lhs_maxval=10.0,
      rhs_maxval=20.0,
      gra_maxval=30.0,
      dims=(((1,), (0,)), ((), ())),  # classical matmul
      lhs_shape=(10, 20),
      rhs_shape=(20, 30),
      gra_shape=(10, 30),  # has to be the shape of the output
      seed=0,
      dtype=jnp.float32,
      clip_gradient=False,
  ):
    # Set use_fwd_quant to None.
    cfg.drhs.rhs.use_fwd_quant = None
    cfg.dlhs.rhs.use_fwd_quant = None
    readonly_cfg = cfg
    del cfg

    lhs = rand_unif(lhs_shape, lhs_maxval, seed, dtype)
    rhs = rand_unif(rhs_shape, rhs_maxval, seed + 1, dtype)
    gra = rand_unif(gra_shape, gra_maxval, seed + 2, dtype)

    # Prepare utility functions for test.
    aqt_dg_full = functools.partial(
        _aqt_dg_full,
        readonly_cfg=readonly_cfg,
        dims=dims,
        clip_gradient=clip_gradient,
    )
    aqt_dg_full_lr_diff = functools.partial(
        _aqt_dg_full_lr_diff,
        readonly_cfg=readonly_cfg,
        dims=dims,
        clip_gradient=clip_gradient,
    )
    aqt_dg_raw = functools.partial(
        _aqt_dg_raw, readonly_cfg=readonly_cfg, dims=dims
    )
    aqt_dg_raw_lr_diff = functools.partial(
        _aqt_dg_raw_lr_diff, readonly_cfg=readonly_cfg, dims=dims
    )
    check = functools.partial(_check_result_eq, lhs=lhs, rhs=rhs, gra=gra)

    # test dot_general
    check([
        (
            "RA L       ",
            aqt_dg_full_lr_diff(
                aqt.DequantMode.OTHER_INPUT,
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(),
        ),
        (
            "RA R       ",
            aqt_dg_full_lr_diff(
                aqt.DequantMode.THIS_INPUT,
                aqt.DequantMode.OTHER_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(),
        ),
        (
            "RA fake    ",
            aqt_dg_full(
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(),
        ),
        (
            "RA L fwd   ",
            aqt_dg_raw_lr_diff(
                aqt.DequantMode.OTHER_INPUT,
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(test_gradient=False),
        ),
        (
            "RA R fwd  ",
            aqt_dg_raw_lr_diff(
                aqt.DequantMode.THIS_INPUT,
                aqt.DequantMode.OTHER_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(test_gradient=False),
        ),
        (
            "RA fake fwd",
            aqt_dg_raw(
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(test_gradient=False),
        ),
    ])

    check([
        (
            "RA fake    ",
            aqt_dg_full(
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                local_aqt=aqt.LocalAqt(2),
            ),
            dict(),
        ),
        (
            "RA L      ",
            aqt_dg_full_lr_diff(
                aqt.DequantMode.OTHER_INPUT,
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
                local_aqt=aqt.LocalAqt(2),
            ),
            dict(),
        ),
        (
            "RA R       ",
            aqt_dg_full_lr_diff(
                aqt.DequantMode.THIS_INPUT,
                aqt.DequantMode.OTHER_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
                local_aqt=aqt.LocalAqt(2),
            ),
            dict(),
        ),
    ])

  def test_dot_general_calibrate_dequant_mode_mismatch(self):
    cfg = config.dot_general_make(8, 8, use_fwd_quant=None)
    dims = (((1,), (0,)), ((), ()))
    lhs = rand_unif((10, 20), 10.0, 0, jnp.float32)
    rhs = rand_unif((20, 30), 20.0, 1, jnp.float32)

    # 1. Raise error when OTHER_INPUT + CONTRACTING_AXIS
    with self.assertRaisesRegex(
        AssertionError,
        ".*Unsupported calibration mode.*dequant mode combination.*",
    ):
      _aqt_dg_full_lr_diff(
          aqt.DequantMode.THIS_INPUT,
          aqt.DequantMode.OTHER_INPUT,
          aqt.CalibrationMode.CONTRACTING_AXIS,
          aqt.CalibrationMode.CONTRACTING_AXIS,
          readonly_cfg=copy.deepcopy(cfg),
          dims=dims,
      )(lhs, rhs)

    # 2. Raise error when OUTPUT + REMAINING_AXIS
    with self.assertRaisesRegex(
        AssertionError,
        ".*Unsupported calibration mode.*dequant mode combination.*",
    ):
      _aqt_dg_full_lr_diff(
          aqt.DequantMode.THIS_INPUT,
          aqt.DequantMode.OUTPUT,
          aqt.CalibrationMode.CONTRACTING_AXIS,
          aqt.CalibrationMode.REMAINING_AXIS,
          readonly_cfg=copy.deepcopy(cfg),
          dims=dims,
      )(lhs, rhs)

  @parameterized.parameters([False, True])
  def test_dot_general_prevent_fwd_quant_with_remaining_axis(
      self, use_fwd_quant
  ):
    """If calibration axis is remaining_axis, use_fwd_quant should be None."""
    cfg = config.dot_general_make(8, 8, use_fwd_quant=use_fwd_quant)
    dims = (((1,), (0,)), ((), ()))
    lhs = rand_unif((10, 20), 10.0, 0, jnp.float32)
    rhs = rand_unif((20, 30), 20.0, 1, jnp.float32)

    with self.assertRaisesRegex(
        AssertionError,
        ".*use_fwd_quant should be set to None.*",
    ):
      _aqt_dg_full_lr_diff(
          aqt.DequantMode.THIS_INPUT,
          aqt.DequantMode.OTHER_INPUT,
          aqt.CalibrationMode.CONTRACTING_AXIS,
          aqt.CalibrationMode.REMAINING_AXIS,
          readonly_cfg=copy.deepcopy(cfg),
          dims=dims,
      )(lhs, rhs)

  @parameterized.parameters([
      dict(cfg=config.dot_general_make(8, 8)),
      dict(cfg=config.dot_general_make(8, 8, dlhs_local_aqt=aqt.LocalAqt(2))),
      dict(cfg=config.dot_general_make(8, 8, drhs_local_aqt=aqt.LocalAqt(2))),
  ])
  def test_dot_general_equality_between_different_calibration_axes(
      self,
      cfg: config.DotGeneral,
  ):
    """Check equality between different calibration axes."""
    dims = (((1,), (0,)), ((), ()))

    # Set use_fwd_quant to None.
    cfg.drhs.rhs.use_fwd_quant = None
    cfg.dlhs.rhs.use_fwd_quant = None
    readonly_cfg = cfg
    del cfg

    # Set the two arguments as powers of 2, to prevent it from having the
    # quantization loss.
    lhs = 2 ** jnp.floor(rand_unif((10, 20), 3.0, 1, jnp.float32) + 3.0)
    rhs = 2 ** jnp.floor(rand_unif((20, 30), 3.0, 2, jnp.float32) + 3.0)
    gra = 2 ** jnp.floor(rand_unif((10, 30), 3.0, 3, jnp.float32) + 3.0)

    # Prepare utility functions for test.
    aqt_dg_full = functools.partial(
        _aqt_dg_full,
        readonly_cfg=readonly_cfg,
        dims=dims
    )
    aqt_dg_full_lr_diff = functools.partial(
        _aqt_dg_full_lr_diff,
        readonly_cfg=readonly_cfg,
        dims=dims
    )
    check = functools.partial(_check_result_eq, lhs=lhs, rhs=rhs, gra=gra)

    # test dot_general
    check([
        (
            "CA         ",
            aqt_dg_full(aqt.DequantMode.OUTPUT, use_fwd_quant=False),
            dict(),
        ),
        (
            "CA fake    ",
            aqt_dg_full(aqt.DequantMode.THIS_INPUT, use_fwd_quant=False),
            dict(),
        ),
        (
            "RA L       ",
            aqt_dg_full_lr_diff(
                aqt.DequantMode.OTHER_INPUT,
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(),
        ),
        (
            "RA R       ",
            aqt_dg_full_lr_diff(
                aqt.DequantMode.THIS_INPUT,
                aqt.DequantMode.OTHER_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(),
        ),
        (
            "RA fake      ",
            aqt_dg_full(
                aqt.DequantMode.THIS_INPUT,
                aqt.CalibrationMode.REMAINING_AXIS,
            ),
            dict(),
        ),
    ])

  def test_dynamic_context(self):
    @jax.jit
    def f(lhs, rhs):
      cfg = config.dot_general_make()
      cfg = config.set_context(cfg, key=jax.random.PRNGKey(4), train_step=None)
      dg = aqt.make_dot_general(cfg)
      return dg(lhs, rhs, (((0,), (0,)), ((), ())))

    lhs, rhs = jnp.array([3.0, 4.0]), jnp.array([4.0, 5.0])
    jax.value_and_grad(f)(lhs, rhs)

  def test_hardware_int8(self, seed=0):
    cfg = config.dot_general_raw_make(8, 8)

    def dg(lhs, rhs):
      ret, _ = cfg(
          lhs,
          rhs,
          None,
          None,
          (((1,), (0,)), ((), ())),
      )
      return ret

    lhs = rand_unif((10, 20), 1.0, seed)
    rhs = rand_unif((20, 30), 1.0, seed + 1)
    test_jaxpr_dtype(lambda: dg(lhs, rhs), [cfg], lhs.dtype)
    assert cfg.lhs.quantizer.numerics.get_dtype() == jnp.int8
    assert cfg.rhs.quantizer.numerics.get_dtype() == jnp.int8

  @parameterized.parameters([
      (1, 1),
      (1, 2),
      (2, 1),
      (2, 2),
      (8, 8),
      (None, 8),
      (8, None),
      (None, None),
  ])
  def test_conv_general_dilated(
      self,
      lhs_bits,
      rhs_bits,
      lhs_maxval=10.0,
      rhs_maxval=20.0,
      seed=0,
  ):
    cfg = config.conv_general_dilated_make(2, lhs_bits, rhs_bits)

    if cfg.lhs:
      # Power-of-2 scales allow FQ and AQT to be exactly the same.
      cfg.lhs.quantizer.po2_scale = True
    if cfg.rhs:
      cfg.rhs.quantizer.po2_scale = True

    batch_n = 10
    contr_n = 20
    feature_n = 30
    lhs = rand_unif((batch_n, 4, 5, contr_n), lhs_maxval, seed)
    rhs = rand_unif((3, 3, contr_n, feature_n), rhs_maxval, seed + 1)

    lax_conv = jax.lax.conv_general_dilated
    aqt_conv_fn = aqt_conv.make_conv_general_dilated(cfg)
    kwargs = {
        "window_strides": (1, 1),
        "padding": "SAME",
        "dimension_numbers": fl._conv_dimension_numbers(lhs.shape),
    }
    lhs_fq = aqt_quantizer.make_fake_quant(cfg.lhs)(lhs)
    rhs_fq = aqt_quantizer.make_fake_quant(cfg.rhs)(rhs)
    prod_fq = lax_conv(lhs_fq, rhs_fq, **kwargs)
    prod_aqt = aqt_conv_fn(lhs, rhs, **kwargs)
    assert (prod_aqt == prod_fq).all()

  @parameterized.parameters([
      dict(
          shard_count=2,
          lhs=[1270.0, 10.0, 1270000.0, 10000.0],
          expected_product=1281280.0,
      ),
      dict(
          shard_count=1,
          lhs=[1270.0, 10.0, 1270000.0, 10000.0],
          expected_product=1280000.0,
      ),
  ])
  def test_local_aqt(self, shard_count, lhs, expected_product):
    # create a config that quantizes both forward and backward passes to int8
    # set the number of shards (local aqt) to 2
    cfg = config.fully_quantized(
        fwd_bits=8,
        bwd_bits=8,
        use_stochastic_rounding=False,
        drhs_local_aqt=aqt.LocalAqt(shard_count),
    )
    cfg.fwd.lhs.quantizer.numerics = cfg.fwd.lhs.quantizer.numerics.replace(
        preserve_max_val=True
    )
    cfg.fwd.rhs.quantizer.numerics = cfg.fwd.rhs.quantizer.numerics.replace(
        preserve_max_val=True
    )
    cfg.drhs.lhs.quantizer.numerics = cfg.drhs.lhs.quantizer.numerics.replace(
        preserve_max_val=True
    )
    cfg.drhs.rhs.quantizer.numerics = cfg.drhs.rhs.quantizer.numerics.replace(
        preserve_max_val=True
    )
    dg = lambda lhs, rhs: aqt.make_dot_general(cfg)(
        lhs,
        rhs,
        dimension_numbers=(((), ()), ((), ())),
    )
    lhs = jnp.array(lhs)
    rhs = jnp.array([1.0])
    output, bprop = jax.vjp(dg, lhs, rhs)
    _, drhs = bprop(jnp.ones_like(output))
    assert drhs == expected_product

  @parameterized.parameters(
      # 'bmnts,bsnh->bmtnh'
      (
          (2, 3, 4, 5, 6),
          (2, 6, 4, 1),
          (((4,), (1,)), ((0, 2), (0, 2))),
          (2, 1, 4, 1, 6),
      ),
      # 'bmkgts,bskh->bmtkgh'
      (
          (2, 3, 4, 5, 6, 7),
          (2, 7, 4, 1),
          (((5,), (1,)), ((0, 2), (0, 2))),
          (2, 1, 4, 1, 1, 7),
      ),
  )
  def test_rhs_scale_transpose_for_lhs_input(
      self, lhs_shape, rhs_scale_shape, dimension_numbers, expected_shape
  ):
    lhs = jnp.zeros(lhs_shape)
    rhs_scale = jnp.zeros(rhs_scale_shape)
    result = aqt._rhs_scale_transpose_for_lhs_input(
        rhs_scale, dimension_numbers, lhs.shape
    )

    self.assertEqual(result.shape, expected_shape)

  @parameterized.parameters(
      # 'bmnts,bsnh->bmtnh'
      (
          (2, 1, 4, 1, 3),
          (2, 3, 4, 5),
          (((4,), (1,)), ((0, 2), (0, 2))),
          (2, 3, 4, 1),
      ),
      # 'bmkgts,bskh->bmtkgh'
      (
          (2, 1, 4, 1, 1, 3),
          (2, 3, 4, 5),
          (((5,), (1,)), ((0, 2), (0, 2))),
          (2, 3, 4, 1),
      ),
  )
  def test_lhs_scale_transpose_for_rhs_input(
      self, lhs_scale_shape, rhs_shape, dimension_numbers, expected_shape
  ):
    lhs_scale = jnp.zeros(lhs_scale_shape)
    rhs = jnp.zeros(rhs_shape)
    result = aqt._lhs_scale_transpose_for_rhs_input(
        lhs_scale, dimension_numbers, rhs.shape
    )

    self.assertEqual(result.shape, expected_shape)

  def test_per_tensor_quant(self):
    x = jnp.array([
        [1, 2, 3, 4],
        [5, 6, 7, 8],
        [3, 5, 1, 127.5],
    ])
    cfg = config.dot_general_make(lhs_bits=8, rhs_bits=8)
    cfg.fwd.lhs.quantizer.calib_shared_axes = "per_tensor"
    # calibration_axes will not be used when cfg.calib_shared_axes is set
    y, _ = cfg.fwd.lhs.quantizer.quant(x, calibration_axes=None)
    self.assertEqual(y.scale[0], jnp.array([[1.0]]))


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Multiple sized model with varied quantization and sparsity.

Quantization is enabled since starting
Sparsity is one shot pruning, updated and applied at 100K steps.
"""

import copy

from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental.tmlr import model_size_sweep
import ml_collections


def get_config():
  """Sweep for multiple sized model with varied quantization and sparsity."""
  sweep_config = ml_collections.ConfigDict()
  configs = []

  for model_config in model_size_sweep.get_model_size_configs():
    for prec, prune_rate in [
        (None, None),
        (8, None),
        (None, (2, 4)),
        (4, None),
        (None, (1, 4)),
        (8, (2, 4)),
        (2, None),
        (None, (1, 8)),
        (4, (2, 4)),
        (8, (1, 4)),
    ]:
      config = copy.deepcopy(model_config)
      config.quant_type = 'aqt'

      # mlp + attention weights + acts  quantization
      config.dense.weight_prec = prec
      config.dense.quant_act.prec = prec

      # mlp weights sparsity
      config.mlp_block.dense_1.weight_sparsity.type = 'STRUCTURED_NM'
      config.mlp_block.dense_1.weight_sparsity.prune_rate = prune_rate
      config.mlp_block.dense_2.weight_sparsity.type = 'STRUCTURED_NM'
      config.mlp_block.dense_2.weight_sparsity.prune_rate = prune_rate
      # attn_weights sparsity
      config.attention.dense_kqv.weight_sparsity.type = 'STRUCTURED_NM'
      config.attention.dense_kqv.weight_sparsity.prune_rate = prune_rate
      config.attention.dense_out.weight_sparsity.type = 'STRUCTURED_NM'
      config.attention.dense_out.weight_sparsity.prune_rate = prune_rate
      config.metadata.hyper_str = (
          f'{config.metadata.hyper_str}_prec({prec})_NM({prune_rate})'
      )
      configs.append(config)

  sweep_config.configs = configs
  return sweep_config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Fork of flax.deprecated.nn.attention modules with optional quantization."""

import dataclasses
import typing
from typing import Any, Callable, Iterable, Optional, Type, TypeVar
from aqt.jax_legacy.jax import flax_layers
from aqt.jax_legacy.jax import get_bounds
from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax import quantization
from aqt.jax_legacy.jax import shape_utils
from aqt.jax_legacy.jax import stats_tag
from aqt.jax_legacy.jax.flax import struct as flax_struct
from aqt.jax_legacy.jax.flax_layers import default_kernel_init
from aqt.jax_legacy.jax.flax_layers import InitializerType
from aqt.jax_legacy.jax.quantization import quantized_dynamic_dot_general
from aqt.jax_legacy.jax.quantization import QuantOps
from aqt.jax_legacy.jax.quantization import QuantType
from flax import linen as nn
# TODO(malmaud): Remove reliance on these 'legacy' nn.attention methods
from flax.deprecated.nn.attention import _make_causal_mask
from flax.deprecated.nn.attention import make_padding_mask
from flax.linen import initializers
import jax
from jax import lax
from jax import random
import jax.numpy as jnp
import numpy as onp

T = TypeVar('T')

dataclass = flax_struct.dataclass if not typing.TYPE_CHECKING else dataclasses.dataclass


@dataclass
class ExpHParams:
  """Hparams to customize exponential."""
  # Limit the total accumulated value of the sum of exponentials
  sum_high_bound: float
  # Defines low bound for (x-m) where x is an element of the incoming vector
  # and m is the max value in the vector.
  low_bound: float
  # Substract exponential evaluated a low bound (low_bound) from calculations so
  # that exponential calculations evaluated at (x-m = low_bound) converges to 0.
  clip_and_subtract: bool
  # Use this gradient as linear approximation of exponential.
  linear_gradient: float


@dataclass
class ReciprocalHParams:
  """Hparams to customize reciprocal."""
  # Use this gradient as linear approximation of exponential.
  linear_gradient: float
  # Defines low bound for input to reciprocal.
  low_bound: float


@dataclass
class SoftmaxQuantHParams:
  """Hyperparameters for quantized softmax operation used in attention."""

  # Floating-point precision to quantize intermediate computations to.
  # If None, intermediate computations will not be quantized.
  prec: Optional[QuantOps.FloatQuant.FloatPrec]

  # Floating-point precision to accumulate the sum reduction in softmax to. If
  # None, sums are accumulated based on the underlying device VPU (float32 for
  # TPU).
  reduction_prec: Optional[QuantOps.FloatQuant.FloatPrec]


@dataclass
class SoftmaxHParams:
  """Hparams to customize softmax."""
  exp_hparams: Optional[ExpHParams]
  reciprocal_hparams: Optional[ReciprocalHParams]
  quant_hparams: Optional[SoftmaxQuantHParams]


@dataclass
class DotProductAttnHParams:
  """HParams class to quantize DotProductAttnAqts."""
  # QuantOps hyperparameter to quantize query for attention weights
  # (i.e. quantizing Q for Q * K).
  attn_act_q: Optional[QuantOps.ActHParams]

  # QuantOps hyperparameter to quantize key for attention weights
  # (i.e. quantizing K for Q * K).
  attn_act_k: Optional[QuantOps.ActHParams]

  # QuantOps hyperparameter to quantize product Q*K (attn_weights).
  attn_act_probs: Optional[QuantOps.ActHParams]

  # QuantOps hyperparameter to quantize value.
  attn_act_v: Optional[QuantOps.ActHParams]

  # Quantization strategy, one of `fake_quant` or `aqt`.
  quant_type: QuantType

  # Custom softmax. Currently we support a linear approximation to exp and
  # reciprocal.
  # We also support downcast intermediate activations to a floating-point
  # format.
  softmax: Optional[SoftmaxHParams]
  # TODO(shivaniagrawal): Changed the strategy to AQT if quant_type is aqt.


def reciprocal(tensor, dtype, recip_hparams: ReciprocalHParams):
  """Generates a reciprocal function based on recip hyper params."""
  if recip_hparams is not None and recip_hparams.linear_gradient != 0:
    # Want: max(low_bound, -a*x+b) such that (-a*x+b) goes through
    # (1, 1)
    # Solution: max(low_bound, a+1- a*x) for arbitrary a>0.
    afull = jnp.full(tensor.shape, recip_hparams.linear_gradient).astype(dtype)
    aplus1full = jnp.full(tensor.shape,
                          1 + recip_hparams.linear_gradient).astype(dtype)
    arecip = jnp.clip(
        lax.sub(aplus1full, lax.mul(afull, tensor)), recip_hparams.low_bound,
        1.).astype(dtype)
  else:
    arecip = lax.reciprocal(tensor)
  return arecip


def exponential(tensor, dtype, exp_hparams: ExpHParams):
  """Calculates an exponential approximation based on exp hyper params."""
  # If low_bound defined, it clips x-M.
  if exp_hparams.low_bound != 0:
    tensor = jnp.clip(tensor, exp_hparams.low_bound, 0.)

  # If clip_and_subtract, replace exp(clip(x-M,low_bound)) term with
  # exp(clip(x-M,low_bound))-exp(low_bound).'
  if exp_hparams.clip_and_subtract:
    tensor = lax.sub(tensor, onp.exp(exp_hparams.low_bound).astype(dtype))
  # If linear_gradient: use this gradient as linear approximation of
  # exponential.
  if exp_hparams.linear_gradient is not None and exp_hparams.linear_gradient != 0:
    # Want: max(0, a*x+b) such that a*x+b goes through (0, 1).
    #
    # This comes out to: max(0, a*x+1), for arbitrary a>0.
    one = jnp.full(tensor.shape, 1.).astype(dtype)
    gradient = jnp.full(tensor.shape, exp_hparams.linear_gradient).astype(dtype)
    approx_exp = jnp.clip(lax.add(lax.mul(tensor, gradient), one), 0, 1)

  else:
    approx_exp = lax.exp(tensor)

  return approx_exp


def softmax(attn_weights, norm_dims, dtype, softmax_hparams: SoftmaxHParams,
            dynamic_context: quant_config.DynamicContext):
  """Normalizes attention."""
  a = attn_weights

  def unquantized_softmax(a):
    a = lax.exp(a -
                jax.scipy.special.logsumexp(a, axis=norm_dims, keepdims=True))
    return a.astype(dtype)

  # Quantize intermediate activations with QuantOps.
  # Currently only supports unscaled floating-point formats.
  def quantized_softmax(a):
    # We compute softmax as exp(x-max(x))/sum_i(exp(x_i-max(x))), quantizing
    # intermediate values. Note this differs from the log-domain
    # implementation of softmax used above.
    quant_hparams = softmax_hparams.quant_hparams
    fp_quant_config = QuantOps.FloatQuant(
        is_scaled=False, fp_spec=quant_hparams.prec)
    quant_ops = QuantOps.create_symmetric_fp(
        fp_quant=fp_quant_config, bounds=None)

    a = quant_ops.to_quantized(a, dtype=dtype)
    # Note that the max of a quantized vector is necessarily also quantized to
    # the same precision since the max of a vector must be an existing element
    # of the vector, so we don't need to explicitly insert a quantization
    # operator to the output of the max reduction.
    a_max = jnp.max(a, axis=norm_dims, keepdims=True)
    a_minus_max = quant_ops.to_quantized(a - a_max, dtype=dtype)
    a_exp = quant_ops.to_quantized(jnp.exp(a_minus_max), dtype=dtype)

    sum_exp_quantized_reduction = quantization.quantized_sum(
        a_exp, axis=norm_dims, keepdims=True, prec=quant_hparams.reduction_prec)
    sum_exp = quant_ops.to_quantized(sum_exp_quantized_reduction, dtype=dtype)

    inv_sum_exp = quant_ops.to_quantized(jnp.reciprocal(sum_exp), dtype=dtype)
    a_softmax = quant_ops.to_quantized(a_exp * inv_sum_exp, dtype=dtype)

    return a_softmax.astype(dtype)

  # If no params, return accurate Softmax.
  if softmax_hparams == SoftmaxHParams(None, None,
                                       None) or softmax_hparams is None:
    return unquantized_softmax(a)

  # TODO(shivaniagrawal): Partial sum quantization (if enabled) will happen for
  # the entire training run, even before the global activation start step.
  if softmax_hparams.quant_hparams is not None:
    return lax.cond(dynamic_context.quantize_acts, quantized_softmax,
                    unquantized_softmax, a)

  # Approximated Softmax
  exp_hparams = softmax_hparams.exp_hparams
  recip_hparams = softmax_hparams.reciprocal_hparams

  # Substract max value from dimensions to be normalized.
  shape = jax.util.subvals(onp.shape(a), zip(norm_dims, (1,) * len(norm_dims)))
  dimadd = lambda x: lax.reshape(x, shape)
  # pylint: disable=protected-access
  amax = lax.reduce(a, onp.array(-onp.inf, dtype=a.dtype), lax.max, norm_dims)
  amax = lax.select(lax.is_finite(amax), amax, lax.full_like(amax, 0))
  amax_singletons = dimadd(amax)
  asubmax = lax.sub(a, amax_singletons)

  # Calculate approximated exponential
  approx_exp = exponential(asubmax, dtype, exp_hparams)

  # If sum_high_bound: Upper clip bound for sum(exp(x-M)).
  asumexp = dimadd(
      lax.reduce(approx_exp, onp.array(0, dtype=a.dtype), lax.add,
                 norm_dims))

  if exp_hparams.sum_high_bound is not None and exp_hparams.sum_high_bound != 0:
    sum_low_bound = 1.
    if (exp_hparams.low_bound != 0) and exp_hparams.clip_and_subtract:
      sum_low_bound = 1 - onp.exp(exp_hparams.low_bound)
    asumexp = jnp.clip(asumexp, sum_low_bound, exp_hparams.sum_high_bound)

  # Approximation of reciprocal.
  arecip = reciprocal(asumexp, dtype, recip_hparams)
  return lax.mul(approx_exp, arecip).astype(dtype)


# Forked from flax.deprecated.nn.attention.dot_product_attention
# https://github.com/google/flax/blob/65061e6128f6695eed441acf2bfffc3b1badd318/flax/nn/attention.py#L206
def dot_product_attention(query,
                          key,
                          value,
                          hparams: DotProductAttnHParams,
                          dynamic_context: quant_config.DynamicContext,
                          paxis_name: Optional[str],
                          train: bool,
                          key_padding_mask: Optional[jnp.ndarray],
                          query_padding_mask: Optional[jnp.ndarray],
                          attn_mask: Optional[jnp.ndarray],
                          dtype=jnp.float32,
                          bias=None,
                          axis=None,
                          broadcast_dropout=True,
                          dropout_rng=None,
                          dropout_rate=0.,
                          deterministic=False,
                          precision=None):
  """Computes dot-product attention given query, key, and value.

  This is the core function for applying attention based on
  https://arxiv.org/abs/1706.03762. It calculates the attention weights given
  query and key and combines the values using the attention weights. This
  function supports multi-dimensional inputs.


  Args:
    query: queries for calculating attention with shape of `[batch_size,
      sequence_length, num_heads, mem_channels]`.
    key: keys for calculating attention with shape of `[batch_size,
      sequence_length, num_heads, mem_channels]`.
    value: values to be used in attention with shape of `[batch_size,
      sequence_length, num_heads, value_channels]`.
    hparams: hyperparameters used for quantization.
    dynamic_context: context for quantization.
    paxis_name: axis_name to which a user `pmaps` the parent module (model),
      refer to jax.pmap() for more documentation. This arg is used for
      get_bounds acts quantization (QuantOps.create_input_fake_quant)
    train: Whether model is training.
    key_padding_mask: boolean mask indicating which elements in 'key' and
      'value' are padding. Must have a shape compatible with 'key' and 'value'.
    query_padding_mask: boolean mask indicating which elements in `query` are
      padding (True means not padding).
    attn_mask: boolean mask indicating which elements of the calculated
      attention weight matrix should be used for collecting activation
      statistics. Should have a shape broadcast-compatible with '[bs,
      sequence_length, sequence_length]'. Must have a shape broadcast-compatible
      'query'.
    dtype: the dtype of the computation (default: float32)
    bias: bias for the attention weights. This can be used for incorporating
      autoregressive mask, padding mask, proximity bias.
    axis: axises over which the attention is applied.
    broadcast_dropout: bool: use a broadcasted dropout along batch dims.
    dropout_rng: JAX PRNGKey: to be used for dropout.
    dropout_rate: dropout rate
    deterministic: bool, deterministic or not (to apply dropout)
    precision: numerical precision of the computation see `jax.lax.Precision`
      for details.

  Returns:
    Output of shape `[bs, sequence_length, num_heads, value_channels]`.
  """
  batch_size, query_sequence_length, num_heads, channel_size = query.shape
  key_sequence_length = key.shape[1]
  shape_utils.assert_shapes_equal(
      key.shape, (batch_size, key_sequence_length, num_heads, channel_size))
  shape_utils.assert_shapes_equal(
      value.shape, (batch_size, key_sequence_length, num_heads, channel_size))
  if key_padding_mask is not None:
    shape_utils.assert_shapes_equal(key_padding_mask.shape,
                                    (batch_size, key_sequence_length, 1, 1))
  if query_padding_mask is not None:
    shape_utils.assert_shapes_equal(query_padding_mask.shape,
                                    (batch_size, query_sequence_length, 1, 1))

  if attn_mask is not None:
    shape_utils.assert_shapes_compatible(
        attn_mask.shape,
        (batch_size, 1, query_sequence_length, key_sequence_length))

  if axis is None:
    axis = tuple(range(1, key.ndim - 2))
  if not isinstance(axis, Iterable):
    axis = (axis,)

  for ax in axis:
    if not (query.ndim >= 3 and 1 <= ax < query.ndim - 2):
      raise ValueError('Attention axis must be between the batch '
                       'axis and the last-two axes.')
  depth = query.shape[-1]
  n = key.ndim
  # batch_dims is  <bs, <non-attention dims>, num_heads>
  batch_dims = tuple(onp.delete(range(n), axis + (n - 1,)))
  # q & k -> (bs, <non-attention dims>, num_heads, <attention dims>, channels)

  qk_perm = batch_dims + axis + (n - 1,)
  key = key.transpose(qk_perm)
  shape_utils.assert_shapes_equal(
      key.shape, (batch_size, num_heads, key_sequence_length, channel_size))

  key_padding_mask_transposed = None
  query_padding_mask_transposed = None
  if key_padding_mask is not None:
    key_padding_mask_transposed = key_padding_mask.transpose(qk_perm)
    shape_utils.assert_shapes_equal(key_padding_mask_transposed.shape,
                                    (batch_size, 1, key_sequence_length, 1))

  if dynamic_context.collect_acts_stats:
    stats_tag.StatsTag(
        channel_axis=None, name='attn_act_k', update_stats=train)(
            key, mask=key_padding_mask_transposed)

  if query_padding_mask is not None:
    query_padding_mask_transposed = query_padding_mask.transpose(qk_perm)
    shape_utils.assert_shapes_equal(query_padding_mask_transposed.shape,
                                    (batch_size, 1, query_sequence_length, 1))

  key_bounds_params = get_bounds.GetBounds.Params(
      update_bounds=dynamic_context.update_bounds,
      update_stats=train,
      paxis_name=paxis_name,
      mask=key_padding_mask_transposed,
      module_name='K')

  # v -> (bs, <non-attention dims>, num_heads, channels, <attention dims>)
  v_perm = batch_dims + (n - 1,) + axis
  value = value.transpose(v_perm)
  shape_utils.assert_shapes_equal(
      value.shape, (batch_size, num_heads, channel_size, key_sequence_length))
  value_padding_mask_transposed = None
  if key_padding_mask is not None:
    value_padding_mask_transposed = key_padding_mask.transpose(v_perm)
    shape_utils.assert_shapes_equal(value_padding_mask_transposed.shape,
                                    (batch_size, 1, 1, key_sequence_length))

  if dynamic_context.collect_acts_stats:
    stats_tag.StatsTag(
        channel_axis=None, name='attn_act_v', update_stats=train)(
            value, mask=value_padding_mask_transposed)

  value_bounds_params = get_bounds.GetBounds.Params(
      update_bounds=dynamic_context.update_bounds,
      update_stats=train,
      paxis_name=paxis_name,
      mask=value_padding_mask_transposed,
      module_name='V')

  query = query / jnp.sqrt(depth).astype(dtype)
  query = query.transpose(qk_perm)
  shape_utils.assert_shapes_equal(
      query.shape, (batch_size, num_heads, query_sequence_length, channel_size))

  if dynamic_context.collect_acts_stats:
    stats_tag.StatsTag(
        channel_axis=None, name='attn_act_q', update_stats=train)(
            query, mask=query_padding_mask_transposed)

  query_bounds_params = get_bounds.GetBounds.Params(
      update_bounds=dynamic_context.update_bounds,
      update_stats=train,
      paxis_name=paxis_name,
      mask=query_padding_mask_transposed,
      module_name='Q')

  batch_dims_t = tuple(range(len(batch_dims)))
  attn_weights = quantized_dynamic_dot_general(
      lhs_act=query,
      rhs_act=key,
      dot_dimension_numbers=(((n - 1,), (n - 1,)), (batch_dims_t,
                                                    batch_dims_t)),
      dot_precision=precision,
      quant_type=hparams.quant_type,
      lhs_act_hparams=hparams.attn_act_q,
      lhs_bounds_params=query_bounds_params,
      rhs_act_hparams=hparams.attn_act_k,
      rhs_bounds_params=key_bounds_params,
  )
  # NOTE(shivaniagrawal): we do per-layer quantization here since that's the
  # only way for activation*activation matmuls to be aqt compatible since we use
  # static scaling factors for activations.

  shape_utils.assert_shapes_equal(
      attn_weights.shape,
      (batch_size, num_heads, query_sequence_length, key_sequence_length))

  # apply attention bias: masking, dropout, proximity bias, ect.
  if bias is not None:
    attn_weights = attn_weights + bias

  # normalize the attention weights
  norm_dims = tuple(range(attn_weights.ndim - len(axis), attn_weights.ndim))
  attn_weights = softmax(
      attn_weights,
      norm_dims,
      dtype,
      hparams.softmax,
      dynamic_context=dynamic_context)

  # apply dropout
  if not deterministic and dropout_rate > 0.0:
    if dropout_rng is None:
      raise ValueError('dropout_rng cannot be None if dropout is requested.')
    keep_prob = 1.0 - dropout_rate
    if broadcast_dropout:
      # dropout is broadcast across the batch+head+non-attention dimension
      dropout_dims = attn_weights.shape[-(2 * len(axis)):]
      dropout_shape = (tuple([1] * len(batch_dims_t)) + dropout_dims)
      keep = random.bernoulli(dropout_rng, keep_prob, dropout_shape)
    else:
      keep = random.bernoulli(dropout_rng, keep_prob, attn_weights.shape)
    multiplier = (
        keep.astype(attn_weights.dtype) / jnp.asarray(keep_prob, dtype=dtype))
    attn_weights = attn_weights * multiplier

  if dynamic_context.collect_acts_stats:
    stats_tag.StatsTag(
        channel_axis=None, name='attn_act_probs', update_stats=train)(
            attn_weights, mask=attn_mask)

  if hparams.attn_act_probs is not None:
    assert hparams.attn_act_probs.bounds == 1.0, (
        'act quantization bounds should '
        'be set to fix value 1.0 to '
        'match Softmax range.')
  probs_bounds_params = get_bounds.GetBounds.Params(
      update_bounds=dynamic_context.update_bounds,
      update_stats=train,
      paxis_name=paxis_name,
      mask=attn_mask,
      module_name='attn_probs')

  # compute the new values given the attention weights
  wv_contracting_dims = (norm_dims, range(value.ndim - len(axis), value.ndim))
  y = quantized_dynamic_dot_general(
      lhs_act=attn_weights,
      rhs_act=value,
      dot_dimension_numbers=(wv_contracting_dims, (batch_dims_t, batch_dims_t)),
      dot_precision=precision,
      quant_type=hparams.quant_type,
      lhs_act_hparams=hparams.attn_act_probs,
      lhs_bounds_params=probs_bounds_params,
      rhs_act_hparams=hparams.attn_act_v,
      rhs_bounds_params=value_bounds_params,
  )
  # NOTE(shivaniagrawal): we do per-layer quantization here since that's the
  # only way for activation*activation matmuls to be aqt compatible since we
  # use static scaling factors for activations.

  shape_utils.assert_shapes_equal(
      y.shape, (batch_size, num_heads, query_sequence_length, channel_size))
  # back to (bs, dim1, dim2, ..., dimN, num_heads, channels)
  perm_inv = _invert_perm(qk_perm)
  y = y.transpose(perm_inv)
  shape_utils.assert_shapes_equal(
      y.shape, (batch_size, query_sequence_length, num_heads, channel_size))
  return y


def _invert_perm(perm):
  perm_inv = [0] * len(perm)
  for i, j in enumerate(perm):
    perm_inv[j] = i
  return tuple(perm_inv)


class MultiHeadDotProductAttentionAqt(nn.Module):
  """Multi-head dot-product attention.

  Attributes:
    num_heads: number of attention heads. Features (i.e. inputs_q.shape[-1])
      should be divisible by the number of heads.
    hparams: hyperparameters
    update_bounds: Bool whether to update activation bounds.
    paxis_name: axis_name to which a user `pmaps` the parent module (model),
      refer to jax.pmap() for more documentation. This arg is used for
      get_bounds acts quantization (QuantOps.create_input_fake_quant)
    train: Whether model is training.
    collect_acts_stats: Whether to tag activations to record statistics.
    dtype: the dtype of the computation (default: float32)
    qkv_features: dimension of the key, query, and value.
    attention_axis: axes over which the attention is applied ( 'None' means
      attention over all axes, but batch, heads, and features).
    causal_mask: boolean specifying whether to apply a causal mask on the
      attention weights. If True, the output at timestep `t` will not depend on
      inputs at timesteps strictly greater than `t`.
    broadcast_dropout: bool: use a broadcasted dropout along batch dims.
    dropout_rate: dropout rate
    deterministic: bool, deterministic or not (to apply dropout)
    kernel_init: initializer for the kernel of the Dense layers.
    bias_init: initializer for the bias of the Dense layers.
    bias: bool: whether pointwise QKVO dense transforms use bias.
    attention_fn: flax.deprecated.nn.dot_product_attention or compatible
      function. Accepts query, key, value, and returns output of shape `[bs,
      dim1, dim2, ..., dimN, num_heads, value_channels]`.
  """

  @dataclass
  class HParams:
    """HParams class to quantize MultiHeadDotProductAttentionAqts."""
    # DenseAQT hyperparameter to quantize key, query and value dense
    # matmuls (i.e. xq * Wq, xk * Wk, xv * Wv) in Attention layer.
    dense_kqv: flax_layers.DenseAqt.HParams
    # DenseAQT hyperparameter to quantize dense_out matmul.
    dense_out: flax_layers.DenseAqt.HParams
    # QuantOps hyperparameter to quantize query, key and value for attention
    # act*act matmauls
    attn_acts: DotProductAttnHParams

  hparams: HParams
  num_heads: int
  paxis_name: Optional[str]
  train: bool
  dynamic_context: quant_config.DynamicContext
  dtype: Type[Any]
  qkv_features: Optional[int]
  attention_axis: Optional[Iterable[int]]
  causal_mask: bool
  dropout_rate: float
  deterministic: bool
  decode: bool
  broadcast_dropout: bool = True
  kernel_init: InitializerType = default_kernel_init
  attention_fn: Callable[..., jnp.ndarray] = dot_product_attention
  bias_init: InitializerType = initializers.zeros
  use_bias: bool = True

  @nn.compact
  def __call__(self,
               inputs_q: jnp.ndarray,
               inputs_kv: jnp.ndarray,
               *,
               padding_mask: Optional[jnp.ndarray],
               key_padding_mask: Optional[jnp.ndarray],
               segmentation: Optional[jnp.ndarray] = None,
               key_segmentation: Optional[jnp.ndarray] = None) -> jnp.ndarray:
    """Applies multi-head dot product attention on the input data.

    If weight_prec is not None, scales and quantizes weights to signed int with
    weight_prec bits.

    Projects the inputs into multi-headed query, key, and value vectors,
    applies dot-product attention and project the results to an output vector.

    This can be used for encoder-decoder attention by specifying both `inputs_q`
    and `inputs_kv` or for self-attention by only specifying `inputs_q` and
    setting `inputs_kv` to None.

    Args:
      inputs_q: input queries of shape `[bs, dim1, dim2, ..., dimN, features]`.
      inputs_kv: key/values of shape `[bs, dim1, dim2, ..., dimN, features]` or
        None for self-attention, inn which case key/values will be derived from
        inputs_q.
      padding_mask: boolean tensor specifying query tokens that are pad token.
      key_padding_mask: boolean tensor specifying key-value tokens that are pad
        token.
      segmentation: segment indices for packed inputs_q data.
      key_segmentation: segment indices for packed inputs_kv data.

    Returns:
      output of shape `[bs, dim1, dim2, ..., dimN, features]`.
    """
    batch_size, query_sequence_length, channel_size = inputs_q.shape
    hparams = self.hparams
    if inputs_kv is None:
      inputs_kv = inputs_q
      key_sequence_length = inputs_q.shape[1]
    else:
      key_sequence_length = inputs_kv.shape[1]
      shape_utils.assert_shapes_equal(
          inputs_kv.shape, (batch_size, key_sequence_length, channel_size))

    jax_precision = jax.lax.Precision.DEFAULT

    if padding_mask is not None:
      shape_utils.assert_shapes_equal(padding_mask.shape,
                                      (batch_size, query_sequence_length, 1))
    if key_padding_mask is None:
      key_padding_mask = padding_mask
    else:
      shape_utils.assert_shapes_equal(key_padding_mask.shape,
                                      (batch_size, key_sequence_length, 1))
    attention_axis = self.attention_axis
    if attention_axis is None:
      attention_axis = tuple(range(1, inputs_q.ndim - 1))

    qkv_features = self.qkv_features
    qkv_features = qkv_features or inputs_q.shape[-1]

    num_heads = self.num_heads
    assert qkv_features % num_heads == 0, (
        'Memory dimension must be divisible by number of heads.')
    head_dim = qkv_features // num_heads

    paxis_name = self.paxis_name
    train = self.train
    kernel_init = self.kernel_init
    bias_init = self.bias_init
    use_bias = self.use_bias
    dtype = self.dtype

    def multi_batch_dense_aqt(inputs, *, name, padding_mask):
      batch_size, sequence_length, channel_size = inputs.shape
      inputs = inputs.reshape(batch_size * sequence_length, channel_size)
      if padding_mask is not None:
        padding_mask = padding_mask.reshape(batch_size * sequence_length, 1)
      out = flax_layers.DenseAqt(
          name=name,
          features=num_heads * head_dim,
          paxis_name=paxis_name,
          train=train,
          dynamic_context=self.dynamic_context,
          hparams=hparams.dense_kqv,
          kernel_init=kernel_init,
          bias_init=bias_init,
          use_bias=use_bias,
          dtype=dtype)(
              inputs, padding_mask=padding_mask)
      return out.reshape(batch_size, sequence_length, num_heads, head_dim)

    # project inputs_q to multi-headed q/k/v
    # dimensions are then [bs, sequence_length, n_heads, n_features_per_head]
    query = multi_batch_dense_aqt(
        inputs_q, name='query', padding_mask=padding_mask)
    key = multi_batch_dense_aqt(
        inputs_kv, name='key', padding_mask=key_padding_mask)
    value = multi_batch_dense_aqt(
        inputs_kv, name='value', padding_mask=key_padding_mask)
    is_cache_initialized = False
    if self.decode:
      is_cache_initialized = self.has_variable('cache', 'cached_key')
      cached_key = self.variable('cache', 'cached_key', jnp.zeros, key.shape,
                                 key.dtype)
      cached_value = self.variable('cache', 'cached_value', jnp.zeros,
                                   value.shape, value.dtype)
      cache_index = self.variable('cache', 'cache_index',
                                  lambda: jnp.array(0, dtype=jnp.int32))
      if is_cache_initialized:
        expected_shape = list(cached_key.value.shape[:-2])
        for attn_dim in attention_axis:
          expected_shape[attn_dim] = 1
        expected_shape = tuple(expected_shape) + inputs_q.shape[-1:]
        if expected_shape != inputs_q.shape:
          raise ValueError('Invalid shape provided, '
                           'expected shape %s instead got %s.' %
                           (expected_shape, inputs_q.shape))

        cshape = cached_key.value.shape
        indices = [0] * len(cshape)
        i = cache_index.value
        attn_size = onp.prod(onp.take(cshape, attention_axis))

        *batch_dims, max_length, num_heads, depth_per_head = (  # pylint: disable=unused-variable
            cached_key.value.shape)
        indices = (0,) * len(batch_dims) + (i, 0, 0)

        key = lax.dynamic_update_slice(cached_key.value, key, indices)
        value = lax.dynamic_update_slice(cached_value.value, value, indices)
        one = jnp.array(1, jnp.int32)
        cache_index.value = cache_index.value + one
        cached_key.value = key
        cached_value.value = value

        # TODO(levskaya): verify this is still needed in translation decoding.
        key_padding_mask = jnp.broadcast_to(
            (jnp.arange(max_length) < cache_index.value), cshape[:2])
        key_padding_mask = key_padding_mask.astype(jnp.float32)[..., None]

    # create attention masks
    mask_components = []
    if self.causal_mask:
      if self.decode and is_cache_initialized:
        bias_pre_shape = (1,) * (key.ndim - 1)
        attn_shape = tuple(onp.take(key.shape, attention_axis))
        attn_size = onp.prod(attn_shape)
        ii = jnp.arange(attn_size, dtype=jnp.int32)
        mask = ii < cache_index.value
        mask_components.append(mask.reshape(bias_pre_shape + attn_shape))
      else:
        mask_components.append(_make_causal_mask(key, attention_axis))
    if padding_mask is not None:
      if key_padding_mask is None:
        key_padding_mask = padding_mask
      attn_padding_mask = make_padding_mask(
          padding_mask_query=padding_mask,
          padding_mask_key=key_padding_mask,
          query_shape=query.shape,
          key_shape=key.shape,
          attention_axis=attention_axis)
      mask_components.append(attn_padding_mask)
    if segmentation is not None:
      if key_segmentation is None:
        key_segmentation = segmentation
      segmentation_mask = make_padding_mask(
          padding_mask_query=segmentation,
          padding_mask_key=key_segmentation,
          query_shape=query.shape,
          key_shape=key.shape,
          attention_axis=attention_axis,
          segmentation_mask=True)
      mask_components.append(segmentation_mask)
    attention_mask = None
    if mask_components:
      attention_mask = mask_components[0]
      for component in mask_components[1:]:
        attention_mask = jnp.logical_and(attention_mask, component)
      attention_mask = attention_mask.astype(jnp.bool_)

      # attention mask in the form of attention bias
      attention_bias = jnp.where(
          attention_mask,
          jnp.full(attention_mask.shape, 0.).astype(dtype),
          jnp.full(attention_mask.shape, -1e10).astype(dtype))
    else:
      attention_bias = None

    # Add an extra dimension to the mask corresponding to the head
    # dimension. eg, if inputs_q has shape [batch_size, sequence_length,
    # n_features], then padding_mask will have a shape
    # [batch_size, sequence_length, 1] and query will have shape
    # [batch_size, sequence_length, n_heads, n_features_per_head].
    # We create query_padding_mask with shape [batch_size, sequence_length,
    # 1, 1] to be broadcast-compatible with 'query'.
    if padding_mask is not None:
      padding_mask = padding_mask[..., None]
      shape_utils.assert_shapes_equal(padding_mask.shape,
                                      (batch_size, query_sequence_length, 1, 1))
    if key_padding_mask is not None:
      key_padding_mask = key_padding_mask[..., None]
      # During prediction, the key padding mask is only going to be
      # broadcast-compatible with the key.
      shape_utils.assert_shapes_compatible(
          key_padding_mask.shape, (batch_size, key_sequence_length, 1, 1))

    # apply attention
    attention_fn = self.attention_fn
    dropout_rate = self.dropout_rate
    broadcast_dropout = self.broadcast_dropout
    deterministic = self.deterministic
    if not deterministic and self.dropout_rate > 0.0:
      dropout_rng = self.make_rng('dropout')
    else:
      dropout_rng = None
    x = attention_fn(  # pylint: disable=redundant-keyword-arg
        query=query,
        key=key,
        value=value,
        hparams=hparams.attn_acts,
        paxis_name=paxis_name,
        train=train,
        dynamic_context=self.dynamic_context,
        dtype=dtype,
        axis=attention_axis,
        bias=attention_bias,
        precision=jax_precision,
        dropout_rng=dropout_rng,
        dropout_rate=dropout_rate,
        broadcast_dropout=broadcast_dropout,
        deterministic=deterministic,
        query_padding_mask=padding_mask,
        key_padding_mask=key_padding_mask,
        attn_mask=attention_mask)
    shape_utils.assert_shapes_equal(
        x.shape, (batch_size, query_sequence_length, num_heads, head_dim))
    x = x.reshape(batch_size * query_sequence_length, num_heads * head_dim)
    if padding_mask is not None:
      padding_mask = padding_mask.reshape(batch_size * query_sequence_length, 1)
    # back to the original inputs dimensions
    out = flax_layers.DenseAqt(
        features=channel_size,
        hparams=hparams.dense_out,
        dynamic_context=self.dynamic_context,
        paxis_name=paxis_name,
        train=train,
        kernel_init=kernel_init,
        bias_init=bias_init,
        use_bias=use_bias,
        dtype=dtype,
        name='dense_out')(
            x, padding_mask=padding_mask)
    shape_utils.assert_shapes_equal(
        out.shape, (batch_size * query_sequence_length, channel_size))
    out = out.reshape(batch_size, query_sequence_length, channel_size)
    return out


class SelfAttentionAqt(MultiHeadDotProductAttentionAqt):
  """Self-attention."""

  @nn.compact
  def __call__(  # pytype: disable=signature-mismatch  # overriding-parameter-count-checks
      self,
      inputs_q: jnp.ndarray,
      *,
      padding_mask: Optional[jnp.ndarray],
      segmentation: Optional[jnp.ndarray] = None,
  ) -> jnp.ndarray:
    return super().__call__(
        inputs_q=inputs_q,
        inputs_kv=inputs_q,
        padding_mask=padding_mask,
        key_padding_mask=padding_mask,
        segmentation=segmentation,
    )



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for aqt.jax.compute_cost_utils."""

import logging

from absl import flags
from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.jax import compute_cost_utils
from aqt.jax_legacy.jax import flax_layers as aqt_flax_layers
from aqt.jax_legacy.jax import get_bounds
from aqt.jax_legacy.jax import hlo_utils
from aqt.jax_legacy.jax import quant_config
from aqt.jax_legacy.jax import quantization
from aqt.jax_legacy.jax.quantization import QuantOps
from aqt.jax_legacy.jax.quantization import QuantType
from flax import linen as nn
from jax import random
from jax._src.lax import convolution as lax_convolution
from jax._src.lax import lax
from jax.nn import initializers
import jax.numpy as jnp
import numpy as onp

FLAGS = flags.FLAGS


class ComputeCostUtilsTest(parameterized.TestCase):

  def setUp(self):
    super(ComputeCostUtilsTest, self).setUp()
    self.rng_key = random.PRNGKey(0)

  def compare_hlo_instructions(self, hlo_no_annotation, hlo_w_annotation):
    """Compares two HLO models to check if they only differ in metadata info."""
    instrs_n = []
    instrs_w = []
    # gather instructions from both HLO models
    for computation in hlo_no_annotation.computations:
      for instr in computation.instructions:
        instrs_n.append(instr)
    for computation in hlo_w_annotation.computations:
      for instr in computation.instructions:
        instrs_w.append(instr)

    self.assertEqual(len(instrs_n), len(instrs_w))
    for i, _ in enumerate(instrs_n):
      # check instructions with the opcode 'convolution'
      # the metadata field for instrs_w and instrs_n should be different.
      if (instrs_n[i].opcode == 'convolution' and
          instrs_w[i].opcode == 'convolution'):
        self.assertNotEqual(instrs_n[i].metadata, instrs_w[i].metadata)

      # remove metadata op_type and op_name
      instrs_n[i].metadata.op_type = ''
      instrs_w[i].metadata.op_type = ''
      instrs_n[i].metadata.op_name = ''
      instrs_w[i].metadata.op_name = ''
      # compare the rest of the instructions.
      self.assertEqual(instrs_n[i], instrs_w[i])

  class TestModelWith1Dense(nn.Module):
    """Test model with a single DenseAqt layer."""

    @nn.compact
    def __call__(self, inputs, hparams, num_classes, dtype=jnp.float32):
      output = aqt_flax_layers.DenseAqt(
          features=num_classes,
          dtype=dtype,
          train=False,
          dynamic_context=quant_config.DynamicContext(
              update_bounds=False, collect_acts_stats=False),
          paxis_name='batch',
          hparams=hparams,
      )(inputs, padding_mask=None)
      return output

  class TestModelWith1Conv(nn.Module):
    """Test model with a single ConvAqt layer."""

    @nn.compact
    def __call__(self,
                 inputs,
                 hparams,
                 kernel_size,
                 num_filters,
                 strides,
                 dtype=jnp.float32):
      output = aqt_flax_layers.ConvAqt(
          features=num_filters,
          kernel_size=kernel_size,
          strides=strides,
          use_bias=False,
          dtype=dtype,
          train=False,
          dynamic_context=quant_config.DynamicContext(update_bounds=False),
          paxis_name='batch',
          hparams=hparams)(
              inputs)
      return output

  class TestModelWith1DynamicMatmul(nn.Module):
    """Test model with a single dynamic matmul."""

    @nn.compact
    def __call__(self, lhs_act, rhs_act, lhs_prec, rhs_prec):
      get_bounds_hyper = get_bounds.GetBounds.Hyper(
          initial_bound=10.0,
          stddev_coeff=0,
          absdev_coeff=0,
          mix_coeff=0,
          granularity=quant_config.QuantGranularity.PER_TENSOR)
      lhs_act_hparams = QuantOps.ActHParams(
          input_distribution='symmetric',
          bounds=get_bounds_hyper,
          prec=lhs_prec,
          half_shift=False)
      rhs_act_hparams = QuantOps.ActHParams(
          input_distribution='symmetric',
          bounds=get_bounds_hyper,
          prec=rhs_prec,
          half_shift=False)
      lhs_bounds_params = get_bounds.GetBounds.Params(
          update_stats=False, update_bounds=False, module_name='lhs')
      rhs_bounds_params = get_bounds.GetBounds.Params(
          update_stats=False, update_bounds=False, module_name='rhs')
      output = quantization.quantized_dynamic_dot_general(
          lhs_act=lhs_act,
          rhs_act=rhs_act,
          lhs_act_hparams=lhs_act_hparams,
          rhs_act_hparams=rhs_act_hparams,
          dot_dimension_numbers=(((1,), (0,)), ((), ())),
          quant_type=QuantType.AQT,
          lhs_bounds_params=lhs_bounds_params,
          rhs_bounds_params=rhs_bounds_params)
      return output

  @parameterized.named_parameters(
      # TestModelWith1Dense
      dict(
          testcase_name='single_dense_layer_bfloat16',
          modelclass=TestModelWith1Dense,
          input_shapes=[(1, 8)],
          model_kwargs={
              'num_classes': 2,
              'hparams': aqt_flax_layers.DenseAqt.HParams(
                  weight_prec=None,
                  quant_type=QuantType.FAKE_QUANT,
                  quant_act=None,
                  weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL,
                  weight_half_shift=False
              ),
          },
          expected_compute_cost=8 * 2 * (16 * 16),
          expected_compute_cost_ratio=1.0,
          expected_compute_cost_linear=8 * 2 * (16),
          expected_compute_cost_ratio_linear=1.0,
          expected_memory_cost=8 * 2 * (16),
          expected_memory_cost_ratio=1.0,
      ),
      dict(
          testcase_name='single_dense_layer_w8_a8',
          modelclass=TestModelWith1Dense,
          input_shapes=[(1, 8)],
          model_kwargs={
              'num_classes': 2,
              'hparams': aqt_flax_layers.DenseAqt.HParams(
                  weight_prec=8,
                  quant_type=QuantType.FAKE_QUANT,
                  quant_act=QuantOps.ActHParams(
                      input_distribution=QuantOps.ActHParams.InputDistribution.POSITIVE,
                      prec=8,
                      bounds=1.0,
                      half_shift=False,
                  ),
                  weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL,
                  weight_half_shift=False
              ),
          },
          expected_compute_cost=8 * 2 * (8 * 8),
          expected_compute_cost_ratio=0.25,
          expected_compute_cost_linear=8 * 2 * (8),
          expected_compute_cost_ratio_linear=0.5,
          expected_memory_cost=8 * 2 * (8),
          expected_memory_cost_ratio=0.5,
      ),

      # TestModelWith1Conv
      dict(
          testcase_name='single_conv_layer_bfloat16',
          modelclass=TestModelWith1Conv,
          input_shapes=[(1, 8, 8, 3)],
          model_kwargs={
              'kernel_size': (3, 3),
              'num_filters': 16,
              'strides': (1, 1),
              'hparams': aqt_flax_layers.ConvAqt.HParams(
                  weight_prec=None,
                  quant_type=QuantType.FAKE_QUANT,
                  quant_act=None,
                  weight_half_shift=False,
              ),
          },
          expected_compute_cost=(3 * 3) * (8 * 8) * 3 * 16 * (16 * 16),
          expected_compute_cost_ratio=1.0,
          expected_compute_cost_linear=(3 * 3) * (8 * 8) * 3 * 16 * (16),
          expected_compute_cost_ratio_linear=1.0,
          expected_memory_cost=(3 * 3) * 3 * 16 * (16),
          expected_memory_cost_ratio=1.0,
      ),
      dict(
          testcase_name='single_conv_layer_bfloat16_strided',
          modelclass=TestModelWith1Conv,
          input_shapes=[(1, 8, 8, 3)],
          model_kwargs={
              'kernel_size': (3, 3),
              'num_filters': 16,
              'strides': (4, 2),
              'hparams': aqt_flax_layers.ConvAqt.HParams(
                  weight_prec=None,
                  quant_type=QuantType.FAKE_QUANT,
                  quant_act=None,
                  weight_half_shift=False,
              ),
          },
          expected_compute_cost=(3 * 3) * ((8 / 4) * (8 / 2)) * 3 * 16 * (16 * 16),
          expected_compute_cost_ratio=1.0,
          expected_compute_cost_linear=(3 * 3) * ((8 / 4) * (8 / 2)) * 3 * 16 * (16),
          expected_compute_cost_ratio_linear=1.0,
          expected_memory_cost=(3 * 3) * 3 * 16 * (16),
          expected_memory_cost_ratio=1.0,
      ),
      dict(
          testcase_name='single_conv_layer_bfloat16_3d',
          modelclass=TestModelWith1Conv,
          input_shapes=[(1, 8, 8, 8, 3)],
          model_kwargs={
              'kernel_size': (3, 3, 3),
              'num_filters': 16,
              'strides': (1, 1, 1),
              'hparams': aqt_flax_layers.ConvAqt.HParams(
                  weight_prec=None,
                  quant_type=QuantType.FAKE_QUANT,
                  quant_act=None,
                  weight_half_shift=False,
              ),
          },
          expected_compute_cost=(3 * 3 * 3) * (8 * 8 * 8) * 3 * 16 * (16 * 16),
          expected_compute_cost_ratio=1.0,
          expected_compute_cost_linear=(3 * 3 * 3) * (8 * 8 * 8) * 3 * 16 * (16),
          expected_compute_cost_ratio_linear=1.0,
          expected_memory_cost=(3 * 3 * 3) * 3 * 16 * (16),
          expected_memory_cost_ratio=1.0,
      ),
      dict(
          testcase_name='single_conv_layer_w4_a2',
          modelclass=TestModelWith1Conv,
          input_shapes=[(1, 8, 8, 3)],
          model_kwargs={
              'kernel_size': (3, 3),
              'num_filters': 16,
              'strides': (1, 1),
              'hparams': aqt_flax_layers.ConvAqt.HParams(
                  weight_prec=4,
                  quant_type=QuantType.FAKE_QUANT,
                  quant_act=QuantOps.ActHParams(
                      input_distribution=QuantOps.ActHParams.InputDistribution.POSITIVE,
                      prec=2,
                      bounds=1.0,
                      half_shift=False,
                  ),
                  weight_half_shift=False,
              ),
          },
          expected_compute_cost=(3 * 3) * (8 * 8) * 3 * 16 * (4 * 2),
          expected_compute_cost_ratio=0.03125,
          expected_compute_cost_linear=(3 * 3) * (8 * 8) * 3 * 16 * (4),
          expected_compute_cost_ratio_linear=0.25,
          expected_memory_cost=(3 * 3) * 3 * 16 * (4),
          expected_memory_cost_ratio=0.25,
      ),
      # TestModelWith1DynamicMatmul
      dict(
          testcase_name='single_dynamic_matmul_layer_bfloat16',
          modelclass=TestModelWith1DynamicMatmul,
          input_shapes=[(1, 8), (8, 1)],
          model_kwargs={'lhs_prec': None,
                        'rhs_prec': None},
          expected_compute_cost=8 * (16 * 16),
          expected_compute_cost_ratio=1.0,
          expected_compute_cost_linear=8 * (16),
          expected_compute_cost_ratio_linear=1.0,
          expected_memory_cost=0,
          expected_memory_cost_ratio=1.0,
      ),
      dict(
          testcase_name='single_dynamic_matmul_layer_l8_r8',
          modelclass=TestModelWith1DynamicMatmul,
          input_shapes=[(1, 8), (8, 1)],
          model_kwargs={'lhs_prec': 8,
                        'rhs_prec': 8},
          expected_compute_cost=8 * (8 * 8),
          expected_compute_cost_ratio=0.25,
          expected_compute_cost_linear=8 * 8,
          expected_compute_cost_ratio_linear=0.5,
          expected_memory_cost=0,
          expected_memory_cost_ratio=1.0,
      ),
      dict(
          testcase_name='single_dynamic_matmul_layer_l8_r4',
          modelclass=TestModelWith1DynamicMatmul,
          input_shapes=[(1, 8), (8, 1)],
          model_kwargs={'lhs_prec': 8,
                        'rhs_prec': 4},
          expected_compute_cost=8 * (8 * 4),
          expected_compute_cost_ratio=0.125,
          expected_compute_cost_linear=8 * (8),
          expected_compute_cost_ratio_linear=0.5,
          expected_memory_cost=0,
          expected_memory_cost_ratio=1.0,
      ),
  )  # pylint: disable=line-too-long
  def test_estimate_simple_model_cost(
      self, modelclass, input_shapes, model_kwargs, expected_compute_cost,
      expected_compute_cost_ratio, expected_compute_cost_linear,
      expected_compute_cost_ratio_linear, expected_memory_cost,
      expected_memory_cost_ratio):
    module = modelclass()
    input_shapes_with_type = [(sh, jnp.float32) for sh in input_shapes]
    dummy_inputs = [
        jnp.ones(input_shape, dtype=dtype)
        for (input_shape, dtype) in input_shapes_with_type
    ]
    init_state = module.init(random.PRNGKey(0), *dummy_inputs, **model_kwargs)

    hlo_proto = hlo_utils.load_hlo_proto_from_model(module, init_state,
                                                    input_shapes,
                                                    **model_kwargs)
    compute_result = compute_cost_utils.estimate_compute_cost(hlo_proto)
    memory_result = compute_cost_utils.estimate_memory_cost(hlo_proto)
    logging.info('compute cost result is %s', compute_result)
    logging.info('memory cost result is %s', memory_result)
    self.assertEqual(compute_result['compute_cost'], expected_compute_cost)
    self.assertEqual(memory_result['memory_cost'], expected_memory_cost)
    self.assertEqual(compute_result['compute_cost_ratio_to_bfloat16'],
                     expected_compute_cost_ratio)
    self.assertEqual(memory_result['memory_cost_ratio_to_bfloat16'],
                     expected_memory_cost_ratio)
    self.assertEqual(compute_result['compute_cost_linear'],
                     expected_compute_cost_linear)
    self.assertEqual(compute_result['compute_cost_ratio_to_bfloat16_linear'],
                     expected_compute_cost_ratio_linear)

  @parameterized.named_parameters(
      # TestModelWith1Dense
      dict(
          testcase_name='single_dense_layer_bfloat16_batch_size',
          modelclass=TestModelWith1Dense,
          input_shape_per_sample=(16,),
          model_kwargs={
              'num_classes':
                  20,
              'hparams':
                  aqt_flax_layers.DenseAqt.HParams(
                      weight_prec=None,
                      quant_act=None,
                      quant_type=QuantType.FAKE_QUANT,
                      weight_quant_granularity=quant_config.QuantGranularity
                      .PER_CHANNEL,
                      weight_half_shift=False)
          },
      ),
      # TestModelWith1Conv
      dict(
          testcase_name='single_conv_layer_bfloat16_batch_size',
          modelclass=TestModelWith1Conv,
          input_shape_per_sample=(16, 16, 3),
          model_kwargs={
              'kernel_size': (3, 3),
              'num_filters':
                  16,
              'strides': (2, 2),
              'hparams':
                  aqt_flax_layers.ConvAqt.HParams(
                      weight_prec=None,
                      quant_act=None,
                      quant_type=QuantType.FAKE_QUANT,
                      weight_half_shift=False,
                  )
          },
      ),
  )
  def test_batch_size_has_no_effect_on_cost(self, modelclass,
                                            input_shape_per_sample,
                                            model_kwargs):
    expected_compute_cost = None
    expected_memory_cost = None
    batch_size_list = [32, 64, 128, 256, 512, 1024]

    module = modelclass()

    # Sweep over the batch size list
    for batch_size in batch_size_list:
      input_shape = (batch_size,) + input_shape_per_sample
      init_state = module.init(
          random.PRNGKey(0), jnp.ones(input_shape, jnp.float32), **model_kwargs)
      hlo_proto = hlo_utils.load_hlo_proto_from_model(module, init_state,
                                                      [input_shape],
                                                      **model_kwargs)
      del init_state
      compute_result = compute_cost_utils.estimate_compute_cost(hlo_proto)
      memory_result = compute_cost_utils.estimate_memory_cost(hlo_proto)
      # Save the first cost and compare it with the rest
      if expected_compute_cost is None:
        expected_compute_cost = compute_result['compute_cost']
      else:
        self.assertEqual(compute_result['compute_cost'], expected_compute_cost)
      if expected_memory_cost is None:
        expected_memory_cost = memory_result['memory_cost']
      else:
        self.assertEqual(memory_result['memory_cost'], expected_memory_cost)

  @parameterized.named_parameters(
      dict(testcase_name='quant_8bit', weight_prec=8),
      dict(testcase_name='quant_4bit', weight_prec=4),
  )
  def test_check_value_inside_and_outside_of_context_conv_general(
      self, weight_prec):
    original_op_name = 'conv_general_dilated'
    # The 'name' in primitive should change in the context in 'flax_layers'
    # if the context is enabled
    self.assertEqual(original_op_name,
                     lax_convolution.conv_general_dilated_p.name)

    with compute_cost_utils.ConvMetadataMonkeyPatch(
        weight_prec=weight_prec, act_prec=None):
      self.assertNotEqual(original_op_name,
                          lax_convolution.conv_general_dilated_p.name)
    self.assertEqual(original_op_name,
                     lax_convolution.conv_general_dilated_p.name)

  @parameterized.named_parameters(
      dict(testcase_name='quant_8bit', weight_prec=8, acts_prec=8),
      dict(testcase_name='quant_4bit', weight_prec=4, acts_prec=4),
  )
  def test_annotation_only_changes_hlo_metadata_conv(self, weight_prec,
                                                     acts_prec):
    FLAGS.metadata_enabled = False
    quant_act = quantization.QuantOps.ActHParams(
        input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
        prec=acts_prec,
        bounds=1.0,
        half_shift=False)
    input_shape = (1, 8, 8, 3)
    module_no_annotation = aqt_flax_layers.ConvAqt(
        features=4,
        kernel_size=(3, 3),
        padding='VALID',
        paxis_name='batch',
        dynamic_context=quant_config.DynamicContext(update_bounds=False),
        train=False,
        hparams=aqt_flax_layers.ConvAqt.HParams(
            weight_prec=weight_prec,
            quant_act=quant_act,
            quant_type=QuantType.FAKE_QUANT,
            weight_half_shift=False),
        kernel_init=initializers.ones,
        bias_init=initializers.ones,
        dtype=jnp.float32)

    init_state = module_no_annotation.init(self.rng_key,
                                           jnp.ones(input_shape, jnp.float32))
    output_no_annotation = module_no_annotation.apply(init_state,
                                                      jnp.ones(input_shape))

    hlo_no_annotation = hlo_utils.load_hlo_proto_from_model(
        module_no_annotation, init_state, [input_shape])
    del init_state

    FLAGS.metadata_enabled = True
    module_w_annotation = aqt_flax_layers.ConvAqt(
        features=4,
        kernel_size=(3, 3),
        padding='VALID',
        paxis_name='batch',
        dynamic_context=quant_config.DynamicContext(update_bounds=False),
        train=False,
        hparams=aqt_flax_layers.ConvAqt.HParams(
            weight_prec=weight_prec,
            quant_act=quant_act,
            quant_type=QuantType.FAKE_QUANT,
            weight_half_shift=False),
        kernel_init=initializers.ones,
        bias_init=initializers.ones,
        dtype=jnp.float32)

    init_state = module_w_annotation.init(self.rng_key,
                                          jnp.ones(input_shape, jnp.float32))
    output_w_annotation = module_w_annotation.apply(init_state,
                                                    jnp.ones(input_shape))

    hlo_w_annotation = hlo_utils.load_hlo_proto_from_model(
        module_w_annotation, init_state, [input_shape])
    del init_state

    onp.testing.assert_array_equal(output_no_annotation, output_w_annotation)
    self.compare_hlo_instructions(hlo_no_annotation, hlo_w_annotation)

  @parameterized.named_parameters(
      dict(testcase_name='quant_8bit', weight_prec=8),
      dict(testcase_name='quant_4bit', weight_prec=4),
  )
  def test_check_value_inside_and_outside_of_context_dot_general(
      self, weight_prec):
    original_op_name = 'dot_general'
    # The 'name' in primitive should change in the context in 'flax_layers'
    # if the context is enabled.
    self.assertEqual(original_op_name, lax.dot_general_p.name)

    with compute_cost_utils.DotMetadataMonkeyPatch(
        lhs_prec=None, rhs_prec=weight_prec, rhs_is_weight=True):
      self.assertNotEqual(original_op_name, lax.dot_general_p.name)
    self.assertEqual(original_op_name, lax.dot_general_p.name)

  @parameterized.named_parameters(
      dict(
          testcase_name='quant_8bit',
          weight_prec=8,
          acts_prec=8,
      ),)
  def test_annotation_only_changes_hlo_metadata_dense(self, weight_prec,
                                                      acts_prec):
    FLAGS.metadata_enabled = False
    quant_act = quantization.QuantOps.ActHParams(
        input_distribution=QuantOps.ActHParams.InputDistribution.SYMMETRIC,
        prec=acts_prec,
        bounds=1.0,
        half_shift=False)
    input_shape = (1, 16)
    module_no_annotation = aqt_flax_layers.DenseAqt(
        features=4,
        use_bias=False,
        dynamic_context=quant_config.DynamicContext(
            update_bounds=False, collect_acts_stats=False),
        paxis_name='batch',
        train=False,
        hparams=aqt_flax_layers.DenseAqt.HParams(
            weight_prec=weight_prec,
            quant_act=quant_act,
            quant_type=QuantType.FAKE_QUANT,
            weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL,
            weight_half_shift=False),
        dtype=jnp.float32)

    init_state = module_no_annotation.init(
        self.rng_key, jnp.ones(input_shape, jnp.float32), padding_mask=None)
    output_no_annotation = module_no_annotation.apply(
        init_state, jnp.ones(input_shape), padding_mask=None)

    hlo_no_annotation = hlo_utils.load_hlo_proto_from_model(
        module_no_annotation, init_state, [input_shape], padding_mask=None)
    del init_state

    FLAGS.metadata_enabled = True
    module_w_annotation = aqt_flax_layers.DenseAqt(
        features=4,
        use_bias=False,
        paxis_name='batch',
        train=False,
        dynamic_context=quant_config.DynamicContext(
            update_bounds=False, collect_acts_stats=False),
        dtype=jnp.float32,
        hparams=aqt_flax_layers.DenseAqt.HParams(
            weight_prec=weight_prec,
            quant_act=quant_act,
            quant_type=QuantType.FAKE_QUANT,
            weight_quant_granularity=quant_config.QuantGranularity.PER_CHANNEL,
            weight_half_shift=False),
    )

    init_state = module_w_annotation.init(
        self.rng_key, jnp.ones(input_shape, jnp.float32), padding_mask=None)
    output_w_annotation = module_w_annotation.apply(
        init_state, jnp.ones(input_shape), padding_mask=None)

    hlo_w_annotation = hlo_utils.load_hlo_proto_from_model(
        module_w_annotation, init_state, [input_shape], padding_mask=None)
    del init_state

    onp.testing.assert_array_equal(output_no_annotation, output_w_annotation)
    self.compare_hlo_instructions(hlo_no_annotation, hlo_w_annotation)


if __name__ == '__main__':
  FLAGS.metadata_enabled = True  # Passes quantization information to HLO
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantization calibration methods."""

import abc
from typing import Union
from aqt.jax.v2 import utils
import jax.numpy as jnp


class Calibration(abc.ABC):

  @abc.abstractmethod
  def get_bound(self, x, shared_axes) -> jnp.ndarray:
    pass


@utils.flax_slots_dataclass
class ConstantCalibration(Calibration):
  bound: Union[jnp.ndarray, float]

  def get_bound(self, x, shared_axes) -> jnp.ndarray:
    """Calibration."""
    del shared_axes
    assert self.bound > 0, 'Bound should be positive.'
    # TODO(yichizh): hardcode bf16 for the scales, subject to quality evaluation
    return jnp.asarray(self.bound).reshape((1,) * len(x.shape)).astype(x.dtype)


@utils.flax_slots_dataclass
class AbsMaxCalibration(Calibration):
  """Simple max(abs(x)) calibration.

  Attributes:
    scale: Set it to something like 0.3, 0.1, 0.03. If scale < 1.0, setting
      IntNumerics.clip_gradient=True is likely to be important.
  """

  scale: float | None = None

  def get_bound(self, x, shared_axes) -> jnp.ndarray:
    """Calibration."""
    msg = (
        'Perhaps you are using DequantMode.THIS_INPUT (fake_quant) and forgot'
        ' to set them.'
    )
    assert shared_axes is not None, msg

    # NOTE: If you want to clip, consider using clip and clip_gradient in
    # int_numerics.IntNumerics.
    abs_max = jnp.max(jnp.abs(x), axis=shared_axes, keepdims=True)
    abs_max = jnp.where(abs_max == 0.0, jnp.ones_like(abs_max), abs_max)
    if self.scale is not None:
      abs_max = abs_max * self.scale
    return abs_max.astype(x.dtype)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""setup.py for AQT."""

import os
from setuptools import find_packages
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))
try:
  README = open(os.path.join(here, "README.md"), encoding="utf-8").read()
except IOError:
  README = ""

jax_legacy_install_requires = [
    "numpy>=1.16.4",
    "jax>=0.3.4",
    "flax>=0.4.1",
    "tensorflow>=2.3.1",
    "tensorflow-datasets>=4.5.2",
    "dacite>=1.6.0",
    "ml_collections>=0.1.1",
]

setup(
    name="aqtp",
    version="0.1.0",
    description="AQT: Accurate Quantized Training",
    long_description="\n\n".join([README]),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    keywords="",
    author="Cerebra Catalyst team",
    author_email="cerebra-catalyst-team@google.com",
    url="https://github.com/google/aqt",
    packages=find_packages(),
    zip_safe=False,
    extras_require={
        "jax_legacy": jax_legacy_install_requires,
    },
)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Different sparsity and quantization strategies for dense activations in small-sized model."""

import copy

from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental.sparsity_and_quantization import baseline
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental.sparsity_and_quantization.dense_acts import quant_4bit
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental.sparsity_and_quantization.dense_acts import quant_8bit
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental.sparsity_and_quantization.dense_acts import quant_8bit_sparsity_structured_2_4
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental.sparsity_and_quantization.dense_acts import sparsity_structured_1_4
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental.sparsity_and_quantization.dense_acts import sparsity_structured_2_4
import ml_collections


def get_config():
  """Returns sweep configuration (see module docstring)."""
  sweep_config = ml_collections.ConfigDict()
  base_configs = [
      baseline,
      quant_4bit,
      sparsity_structured_1_4,
      quant_8bit_sparsity_structured_2_4,
      sparsity_structured_2_4,
      quant_8bit,
  ]
  configs = []
  for base_config_file in base_configs:
    base_config = base_config_file.get_config()
    config = copy.deepcopy(base_config)
    config.metadata.hyper_str = f"{config.metadata.hyper_str}"
    configs.append(config)
  sweep_config.configs = configs
  return sweep_config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flax layer for AQT injection."""

# pylint: disable=unnecessary-lambda
# pylint: disable=g-importing-member
import copy
import functools
from typing import Iterable
from typing import Optional, Union
from aqt.jax.v2 import aqt_dot_general
from aqt.jax.v2 import aqt_tensor
from aqt.jax.v2 import config
from aqt.jax.v2 import tiled_dot_general
from aqt.jax.v2.flax.utils import QuantMode
import flax.linen as nn
import jax
import jax.numpy as jnp


class Freezer(nn.Module):
  """Identity function that can freeze its input.

  On default it is an identity function that saves the input in a variable.
  In 'quant_mode=QuantMode.Serve' mode, ignores the input and returns the frozen
  value. It is usefult to implement 'constant folding' and put quantized weights
  and scales in the checkpoint for serving. Specifically:

  self.get() returns None when quant_mode=TRAIN or CONVERT, returns variable
  when quant_mode=SERVE.
  self.set() does nothing when quant_mode=TRAIN or SERVE, creates and stores
  quantized tensor when quant_mode=CONVERT.
  """

  quant_collection: str
  quant_mode: QuantMode
  q_shape: Iterable[int]
  q_dtype: jnp.dtype
  q_init: nn.initializers.Initializer
  s_shape: Iterable[int]
  s_init: nn.initializers.Initializer

  def setup(self):
    mode = self.quant_mode
    if mode == QuantMode.SERVE or mode == QuantMode.CONVERT:
      collection = self.quant_collection
      q_init = self.q_init
      q_shape = self.q_shape
      q_dtype = self.q_dtype
      s_init = self.s_init
      s_shape = self.s_shape
      # TODO(lew): Store whole QTensor?
      # We could have created one self.variable whose value is a QTensor,
      # but we are unsure how this would complicate the init function,
      # which could potentially be used by adding metadata such as
      # sharding axises, etc.
      self.qvalue = self.variable(collection, 'value', q_init, q_shape, q_dtype)
      self.scale_t = self.variable(collection, 'scale', s_init, s_shape)

  def get(self) -> Optional[aqt_tensor.QTensor]:
    if self.quant_mode == QuantMode.TRAIN:
      return None
    elif self.quant_mode == QuantMode.CONVERT:
      return None
    elif self.quant_mode == QuantMode.SERVE:
      qvalue = self.qvalue.value
      # TODO(b/325626080): Remove the optional logic.
      if self.q_dtype == jnp.int4:
        qvalue = qvalue.astype(jnp.int4)
      return aqt_tensor.QTensor(
          qvalue,
          scale=None,
          scale_t=[self.scale_t.value],
          dequant_dtype=None,  # Rely on dg output dtype for dequant
      )
    else:
      assert False, 'Unknown quant mode.'

  def set(self, inputs: aqt_tensor.QTensor) -> None:
    # TODO(b/325626080): Uncomment the assert.
    # assert inputs.qvalue.dtype == self.q_dtype, (
    #     f'Freezer got a QTensor of type {inputs.qvalue.dtype} but expected'
    #     f' {self.q_dtype}.'
    # )
    if self.quant_mode == QuantMode.TRAIN:
      pass
    elif self.quant_mode == QuantMode.CONVERT:
      qvalue = inputs.qvalue
      # TODO(b/325626080): Remove the optional logic.
      if self.q_dtype == jnp.int4:
        assert qvalue.dtype == jnp.int4
        qvalue = qvalue.astype(jnp.int8)
      self.qvalue.value = qvalue
      assert inputs.scale_t is not None and len(inputs.scale_t) == 1
      self.scale_t.value = inputs.scale_t[0]
    elif self.quant_mode == QuantMode.SERVE:
      # TODO(lew): Optionally compare stored and served value.
      pass
    else:
      assert False, 'Unknown quant mode.'
    return None


class AqtDotGeneral(nn.Module):
  """A layer that can be injected into flax.nn.Dense, etc."""
  cfg: Optional[aqt_dot_general.DotGeneral] = None
  prng_name: Optional[str] = 'params'

  # TODO(lew): split out separate class for each side.
  # Quant mode determines whether flax variables are created to store quantized
  # inputs. Refer to the Freezer doc str to see variable creation in each mode.
  lhs_quant_mode: QuantMode = QuantMode.TRAIN
  # apply_quant_mode determines if using Freezer in cfg.get/set_tensor
  lhs_apply_quant_mode: bool = True
  lhs_init: nn.initializers.Initializer = jnp.zeros
  lhs_scale_init: nn.initializers.Initializer = jnp.zeros
  lhs_var_name: str = 'qlhs'
  lhs_qtensor: Optional[aqt_tensor.QTensor] = None

  rhs_quant_mode: QuantMode = QuantMode.TRAIN
  rhs_apply_quant_mode: bool = True
  rhs_init: nn.initializers.Initializer = jnp.zeros
  rhs_scale_init: nn.initializers.Initializer = jnp.zeros
  rhs_var_name: str = 'qrhs'
  rhs_qtensor: Optional[aqt_tensor.QTensor] = None

  # If you want use 'params' make sure that there is another mechanism to hide
  # these variables from the optimizer.
  quant_collection: str = 'aqt'
  tiling_cfg: Optional[tiled_dot_general.Cfg] = None

  def make_aqt_dg(
      self,
      lhs_shape,
      rhs_shape,
      dimension_numbers: tuple[Iterable[int], Iterable[int]],
  ):
    if self.cfg is None:
      return jax.lax.dot_general

    cfg = copy.deepcopy(self.cfg)
    lhs_scale_shape = list(lhs_shape)
    rhs_scale_shape = list(rhs_shape)
    (contr, _) = dimension_numbers
    for li, ri in zip(*contr):
      lhs_scale_shape[li] = 1
      rhs_scale_shape[ri] = 1
    lhs_scale = aqt_dot_general._lhs_scale_transpose_to_output(  # pylint: disable=protected-access
        jnp.zeros(lhs_scale_shape), dimension_numbers, lhs_shape, rhs_shape
    )
    assert lhs_scale is not None
    lhs_scale_shape = lhs_scale.shape
    rhs_scale = aqt_dot_general._rhs_scale_transpose_to_output(  # pylint: disable=protected-access
        jnp.zeros(rhs_scale_shape), dimension_numbers, lhs_shape, rhs_shape
    )
    assert rhs_scale is not None
    rhs_scale_shape = rhs_scale.shape
    rhs_qm = self.rhs_quant_mode
    lhs_qm = self.lhs_quant_mode

    lhs_freezer = Freezer(
        name=self.lhs_var_name,
        quant_mode=lhs_qm,
        q_shape=lhs_shape,
        q_dtype=cfg.fwd.lhs.quantizer.numerics.get_dtype(),
        q_init=self.lhs_init,
        s_shape=lhs_scale_shape,
        s_init=self.lhs_scale_init,
        quant_collection=self.quant_collection,
    )

    rhs_freezer = Freezer(
        name=self.rhs_var_name,
        quant_mode=rhs_qm,
        q_shape=rhs_shape,
        q_dtype=cfg.fwd.rhs.quantizer.numerics.get_dtype(),
        q_init=self.rhs_init,
        s_shape=rhs_scale_shape,
        s_init=self.rhs_scale_init,
        quant_collection=self.quant_collection,
    )

    prng_name = self.prng_name
    key = self.make_rng(prng_name) if prng_name is not None else None
    cfg = config.set_context(cfg, key, train_step=None)

    def ret_dg(
        lhs,
        rhs,
        dimension_numbers,
        precision=None,
        preferred_element_type=None,
    ):
      del preferred_element_type
      assert (
          precision is None
      ), f'Precision {precision} requested together with quantization.'

      # TODO(yichizh): asserting xhs dtype only when apply_quant_mode=False
      # and cfg.get_qtensor() is None
      msg = 'AQT is not yet optimized to accept quantized types directly. '
      msg += f'lhs.dtype: {lhs.dtype}, rhs.dtype: {rhs.dtype}'
      assert lhs.dtype in [jnp.bfloat16, jnp.float32, jnp.float16], msg
      assert rhs.dtype in [jnp.bfloat16, jnp.float32, jnp.float16], msg

      config.assert_config_validity(cfg)

      # Getter
      lhs_apply_quant_mode = self.lhs_apply_quant_mode
      rhs_apply_quant_mode = self.rhs_apply_quant_mode
      lhs_qt = lhs_freezer.get() if lhs_apply_quant_mode else self.lhs_qtensor
      rhs_qt = rhs_freezer.get() if rhs_apply_quant_mode else self.rhs_qtensor
      out, (out_lhs_qt, out_rhs_qt) = cfg.dg_core(
          lhs=lhs,
          rhs=rhs,
          lhs_qt=lhs_qt,
          rhs_qt=rhs_qt,
          dimension_numbers=dimension_numbers,
      )
      # Setter
      if self.lhs_apply_quant_mode:
        lhs_freezer.set(out_lhs_qt)
      if self.rhs_apply_quant_mode:
        rhs_freezer.set(out_rhs_qt)

      return out

    return ret_dg

  @nn.compact
  def __call__(
      self,
      lhs,
      rhs,
      dimension_numbers,
      precision,
      preferred_element_type=None,
  ):

    aqt_dg = self.make_aqt_dg(lhs.shape, rhs.shape, dimension_numbers)
    if self.tiling_cfg is not None:
      # We integrate tiling here and not on Jax level, so that the Freezers
      # observe tiled shapes.
      aqt_dg = functools.partial(
          tiled_dot_general.tiled_dot_general,
          self.tiling_cfg,
          dot_general=aqt_dg,
      )
    return aqt_dg(
        lhs,
        rhs,
        dimension_numbers,
        precision,
        preferred_element_type,
    )


class AqtEinsum(nn.Module):
  """Quantized Einsum class for model injection."""

  cfg: Optional[config.DotGeneral] = None
  prng_name: Optional[str] = 'params'

  # TODO(lew): split out separate class for each side.
  lhs_quant_mode: QuantMode = QuantMode.TRAIN
  lhs_init: nn.initializers.Initializer = jnp.zeros
  lhs_scale_init: nn.initializers.Initializer = jnp.zeros
  lhs_var_name: str = 'qlhs'

  rhs_quant_mode: QuantMode = QuantMode.TRAIN
  rhs_init: nn.initializers.Initializer = jnp.zeros
  rhs_scale_init: nn.initializers.Initializer = jnp.zeros
  rhs_var_name: str = 'qrhs'

  # If you want use 'params' make sure that there is another mechanism to hide
  # these variables from the optimizer.
  quant_collection: str = 'aqt'

  @nn.compact
  def __call__(
      self,
      eqn,
      lhs_g: Union[jnp.ndarray, aqt_tensor.QTensor],
      rhs_g: Union[jnp.ndarray, aqt_tensor.QTensor],
  ):
    cfg = self.cfg
    lhs_is_qt = isinstance(lhs_g, aqt_tensor.QTensor)
    rhs_is_qt = isinstance(rhs_g, aqt_tensor.QTensor)
    msg = 'Aqt config is None but inputs to AqtEinsum are QTensor.'
    assert not ((lhs_is_qt or rhs_is_qt) and cfg is None), msg
    # when inputs are qtensor, xhs_in is a dummy input that will be consumed by
    # lax einsum API, but it is not used for computation in aqt_dg because it
    # will be overwritten by get_tensor()
    # TODO(lew): We can pass QTensor to lax_numpy._einsum if we add some
    # specific methods to QTensor.
    lhs_in = jnp.zeros_like(lhs_g.qvalue) if lhs_is_qt else lhs_g
    rhs_in = jnp.zeros_like(rhs_g.qvalue) if rhs_is_qt else rhs_g

    # Set the types of dummy input to the same as original input, to prevent it
    # from being rejected by assertions in aqt_dot_general.py, line 522-526 and
    # 414.
    # TODO: b/322111904 - Handle this in more proper way.
    lhs_in, rhs_in = nn.dtypes.promote_dtype(lhs_in, rhs_in)

    # yes_swap = whether einsum swaps [lhs,rhs] when passing them to dot_general
    einsum = functools.partial(aqt_dot_general.einsum, eqn=eqn)
    a = jax.make_jaxpr(einsum)(lhs=lhs_in, rhs=rhs_in)
    [lhs_g_id, rhs_g_id] = a.eqns[0].invars
    [lhs_l_id, rhs_l_id] = a.jaxpr.invars
    not_swap = lhs_g_id == lhs_l_id and rhs_g_id == rhs_l_id
    yes_swap = lhs_g_id == rhs_l_id and rhs_g_id == lhs_l_id
    assert not_swap != yes_swap

    prng_name = self.prng_name

    lhs_quant_mode = self.lhs_quant_mode
    lhs_init = self.lhs_init
    lhs_scale_init = self.lhs_scale_init
    lhs_var_name = self.lhs_var_name
    lhs_qtensor = lhs_g if lhs_is_qt else None

    rhs_quant_mode = self.rhs_quant_mode
    rhs_init = self.rhs_init
    rhs_scale_init = self.rhs_scale_init
    rhs_var_name = self.rhs_var_name
    rhs_qtensor = rhs_g if rhs_is_qt else None

    quant_collection = self.quant_collection

    if yes_swap:
      if cfg is not None:
        cfg = copy.deepcopy(cfg)
        cfg.fwd.lhs, cfg.fwd.rhs = cfg.fwd.rhs, cfg.fwd.lhs
        cfg.dlhs, cfg.drhs = cfg.drhs, cfg.dlhs
      lhs_quant_mode, rhs_quant_mode = rhs_quant_mode, lhs_quant_mode
      lhs_init, rhs_init = rhs_init, lhs_init
      lhs_scale_init, rhs_scale_init = rhs_scale_init, lhs_scale_init
      lhs_var_name, rhs_var_name = rhs_var_name, lhs_var_name
      lhs_is_qt, rhs_is_qt = rhs_is_qt, lhs_is_qt
      lhs_qtensor, rhs_qtensor = rhs_qtensor, lhs_qtensor

    aqt_dg = AqtDotGeneral(
        cfg=cfg,
        prng_name=prng_name,
        lhs_quant_mode=lhs_quant_mode,
        # when passing pre-computed qtensor as inputs, apply_quant_mode flag
        # should be set to False so that Freezer will not be set to overwrite
        # the qtensor passed to dg.
        lhs_apply_quant_mode=not lhs_is_qt,  # Freezer not used if lhs is qt
        lhs_init=lhs_init,
        lhs_scale_init=lhs_scale_init,
        lhs_var_name=lhs_var_name,
        lhs_qtensor=lhs_qtensor,
        rhs_quant_mode=rhs_quant_mode,
        rhs_apply_quant_mode=not rhs_is_qt,  # Freezer not used if rhs is qt
        rhs_init=rhs_init,
        rhs_scale_init=rhs_scale_init,
        rhs_var_name=rhs_var_name,
        rhs_qtensor=rhs_qtensor,
        quant_collection=quant_collection,
    )
    return einsum(lhs=lhs_in, rhs=rhs_in, dg=aqt_dg)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Metadata class and utils shared by the TF and JAX implementations."""
import enum


# the bias of exponent in bfloat16/float32.
EXPONENT_BIAS = 127

# Supported rounding mode.
ROUND_AWAY_FROM_ZERO = 'round_away_from_zero'
ROUND_TO_NEAREST_EVEN = 'round_to_nearest_even'
ROUND_STOCHASTIC = 'round_stochastic'


@enum.unique
class EmulatedFromatsEnum(enum.Enum):
  """Enum for FP8 formats with rounding."""
  # E5M2B15 with round-to-nearest-even
  E5M2B15_RTNE = 'e5m2b15_rtne'
  # E5M2B15 with stochastic rounding
  E5M2B15_STOC = 'e5m2b15_stoc'

  # E4M3B11 with round-to-nearest-even
  E4M3B11_RTNE = 'e4m3b11_rtne'
  # E4M3B11 with stochastic rounding
  E4M3B11_STOC = 'e4m3b11_stoc'

  # E4M3B7 with round-to-nearest-even
  E4M3B7_RTNE = 'e4m3b7_rtne'
  # E4M3B7 with stochastic rounding
  E4M3B7_STOC = 'e4m3b7_stoc'

  # The following uses xla reduce precision ops.
  # E8M2B127 with round-to-nearest-even
  E8M2B127_RTNE = 'e8m2b127_rtne'
  # E8M3B127 with round-to-nearest-even
  E8M3B127_RTNE = 'e8m3b127_rtne'
  # E8M4B127 with round-to-nearest-even
  E8M4B127_RTNE = 'e8m4b127_rtne'
  # E8M5B127 with round-to-nearest-even
  E8M5B127_RTNE = 'e8m5b127_rtne'
  # E8M6B127 with round-to-nearest-even
  E8M6B127_RTNE = 'e8m6b127_rtne'

  # INT8 with rounding formats
  # INT8 with round-to-nearest-even
  INT8_RTNE = 'int8_rtne'
  # INT8 with stochastic rounding
  INT8_STOC = 'int8_stoc'

  # Not emulated formats. placed here for comparison.
  BFLOAT16 = 'bfloat16'
  INT8 = 'int8'


class FPMetadata():
  """Defines floating point formats.

  The max_exp and mix_exp are normal exponent. Subnormals are handled
  separately.
  """

  def __init__(self, name, exponent_bits, mantissa_bits, min_exp, max_exp,
               support_inf, rounding_mode):
    self.name = name
    self.exponent_bits = exponent_bits
    self.mantissa_bits = mantissa_bits
    self.min_exp = min_exp
    self.max_exp = max_exp
    self.support_inf = support_inf
    self.rounding_mode = rounding_mode

  name = ''
  exponent_bits = 8
  mantissa_bits = 7
  min_exp = -126
  max_exp = 128
  support_inf = True
  rounding_mode = ROUND_TO_NEAREST_EVEN


def get_metadata(target_format):
  """Return a FPMetadata based on the format."""
  if target_format == EmulatedFromatsEnum.E5M2B15_RTNE:
    return FPMetadata(target_format, 5, 2, -14, 15, True, ROUND_TO_NEAREST_EVEN)
  elif target_format == EmulatedFromatsEnum.E5M2B15_STOC:
    return FPMetadata(target_format, 5, 2, -14, 15, True, ROUND_STOCHASTIC)
  elif target_format == EmulatedFromatsEnum.E4M3B11_RTNE:
    return FPMetadata(target_format, 4, 3, -10, 4, False, ROUND_TO_NEAREST_EVEN)
  elif target_format == EmulatedFromatsEnum.E4M3B11_STOC:
    return FPMetadata(target_format, 4, 3, -10, 4, False, ROUND_STOCHASTIC)
  elif target_format == EmulatedFromatsEnum.E4M3B7_RTNE:
    return FPMetadata(target_format, 4, 3, -6, 8, False, ROUND_TO_NEAREST_EVEN)
  elif target_format == EmulatedFromatsEnum.E4M3B7_STOC:
    return FPMetadata(target_format, 4, 3, -6, 8, False, ROUND_STOCHASTIC)
  elif target_format == EmulatedFromatsEnum.E8M2B127_RTNE:
    return FPMetadata(target_format, 8, 2, -126, 127, True,
                      ROUND_TO_NEAREST_EVEN)
  elif target_format == EmulatedFromatsEnum.E8M3B127_RTNE:
    return FPMetadata(target_format, 8, 3, -126, 127, True,
                      ROUND_TO_NEAREST_EVEN)
  elif target_format == EmulatedFromatsEnum.E8M4B127_RTNE:
    return FPMetadata(target_format, 8, 4, -126, 127, True,
                      ROUND_TO_NEAREST_EVEN)
  elif target_format == EmulatedFromatsEnum.E8M5B127_RTNE:
    return FPMetadata(target_format, 8, 5, -126, 127, True,
                      ROUND_TO_NEAREST_EVEN)
  elif target_format == EmulatedFromatsEnum.E8M6B127_RTNE:
    return FPMetadata(target_format, 8, 6, -126, 127, True,
                      ROUND_TO_NEAREST_EVEN)
  else:
    raise ValueError('Not supported format: ' + target_format)


def is_fp8_format(target_format):
  """Return if the target_format is a supported FP8 format."""
  return target_format in [
      EmulatedFromatsEnum.E5M2B15_RTNE, EmulatedFromatsEnum.E5M2B15_STOC,
      EmulatedFromatsEnum.E4M3B11_RTNE, EmulatedFromatsEnum.E4M3B11_STOC,
      EmulatedFromatsEnum.E4M3B7_RTNE, EmulatedFromatsEnum.E4M3B7_STOC
  ]


def is_e8_format(target_format):
  """Return if the target_format has 8-bit exponent."""
  return get_metadata(target_format).exponent_bits == 8


def get_max_number_from_mantissa_and_max_exp(mantissa_bits, max_exp):
  """Returns the maximum possible number (not inf) allowed by the format."""
  # NOTE(dotzel): max_mantissa_value is always 1.1111111...1 where number of
  # fractional 1s is mantissa_bits. In the limit mantissa_bits => inf, this exp
  # => 2. It only deviates from 2 by its LSB, 2 ** (-mantissa_bits).
  max_mantissa_value = 2 - 2 ** (-mantissa_bits)
  return (2**max_exp) * max_mantissa_value


def get_max_number(target_format):
  """Returns the maximum possible number (not inf) allowed by the format."""
  fp_metadata = get_metadata(target_format)
  return get_max_number_from_mantissa_and_max_exp(
      fp_metadata.mantissa_bits, fp_metadata.max_exp)





# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""utils contains all kinds of small pieces of code used throughout AQT Flax.

Code in this file can't depend on any other AQT file.
However, it is acceptable to grow pieces of funcionality in this file and
later promote them to dedicated files.
"""

import enum


class QuantMode(enum.Enum):
  TRAIN = 1
  CONVERT = 2
  SERVE = 3



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for analysis_utils.py."""

from absl.testing import absltest
from absl.testing import parameterized
from aqt.jax_legacy.utils import analysis_utils
from aqt.jax_legacy.utils import report_utils
import pandas as pd


class AnalysisUtilsTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    # Create a test ExperimentReport instance
    compute_cost_dict = {
        'compute_cost': 4,
        'memory_cost': 5,
    }

    metrics_dict = {
        'eval': {
            'accuracy': 0.6,
        },
        'train': {
            'accuracy': 0.7,
            'loss': 1.3,
        }
    }
    unsmoothed_metrics_dict = {
        'eval': {
            'accuracy': 0.5,
        },
        'train': {
            'accuracy': 0.8,
            'loss': 1.2,
        }
    }


    report_query_args = {
        'early_stop_attr': 'loss',
        'early_stop_agg': 'MIN',
        'early_stop_ds_dir': 'eval',
        'other_ds_dirs': ['train'],
        'tags_to_include': [
            'accuracy',
            'loss',
        ],
        'smoothing_kernel': 'RECTANGULAR',
        'window_size_in_steps': 50,
        'start_step': 100
    }

    self.test_report = report_utils.ExperimentReport(
        model_dir='/your/model_dir',
        metrics=metrics_dict,
        unsmoothed_metrics=unsmoothed_metrics_dict,
        early_stop_step=1500,
        num_train_steps=2000,
        report_query_args=report_query_args,
        experiment_name='test_exp',
        user_name='test_user',
        launch_time='202102016T112344',
        eval_freq=10,
        first_nan_step=None,
        tensorboard_id='12345',
    )

  def test_convert_report_to_flat_dict_default(self):
    res = analysis_utils.convert_report_to_flat_dict_default(self.test_report)
    exp = {
        'experiment_name': 'test_exp',
        'model_dir': '/your/model_dir',
        'user_name': 'test_user',
        'launch_time': '202102016T112344',
        'eval_freq': 10,
        'first_nan_step': None,
        'tensorboard_id': '12345',
        'early_stop_step': 1500,
        'num_train_steps': 2000,
        'eval/accuracy': 0.6,
        'train/accuracy': 0.7,
        'train/loss': 1.3,
        'unsmoothed/eval/accuracy': 0.5,
        'unsmoothed/train/accuracy': 0.8,
        'unsmoothed/train/loss': 1.2,
        'report_query_args': {
            'early_stop_attr': 'loss',
            'early_stop_agg': 'MIN',
            'early_stop_ds_dir': 'eval',
            'other_ds_dirs': ['train'],
            'tags_to_include': [
                'accuracy',
                'loss',
            ],
            'smoothing_kernel': 'RECTANGULAR',
            'window_size_in_steps': 50,
            'start_step': 100
        },
    }
    self.assertDictEqual(res, exp)

  def test_convert_reports_to_dataframe(self):
    res = analysis_utils.convert_reports_to_dataframe(
        [self.test_report, self.test_report])
    exp_df_shape = (2, 20)
    self.assertEqual(res.shape, exp_df_shape)

  @parameterized.named_parameters(
      dict(
          testcase_name='nested_two_levels',
          nested_dict={
              'a': {
                  'foo': 1,
                  'bar': 2
              },
              'b': 'zoo'
          },
          exp={
              'a/foo': 1,
              'a/bar': 2,
              'b': 'zoo'
          }),
      dict(
          testcase_name='nested_three_levels',
          nested_dict={
              'a': {
                  'foo': {
                      'gan': 1,
                  },
                  'bar': 2
              },
              'b': 'zoo'
          },
          exp={
              'a/foo/gan': 1,
              'a/bar': 2,
              'b': 'zoo'
          }),
      dict(
          testcase_name='nested_with_list',
          nested_dict={
              'a': {
                  'foo': 1,
                  'bar': 2
              },
              'b': ['yaa', 'zoo']
          },
          exp={
              'a/foo': 1,
              'a/bar': 2,
              'b/0': 'yaa',
              'b/1': 'zoo',
          }),
  )
  def test_flatten_with_joined_string_paths(self, nested_dict, exp):
    res = analysis_utils.flatten_with_joined_string_paths(nested_dict)
    self.assertDictEqual(res, exp)

  def test_clickable_link(self):
    link = 'http://test.url'
    display_str = 'test_link'
    res = analysis_utils.clickable_link(link, display_str)
    exp = '<a href="http://test.url">test_link</a>'
    self.assertEqual(res, exp)



if __name__ == '__main__':
  absltest.main()
